#include "headers.h"
#include "tape/tape.h"
#include "arg/arg.h"

int triangulate, sorted, plot, debug, randmode, outputmode;
extern int do_extents;
extern int do_blockers;
extern int do_ff;
extern int do_solidangle;
extern int draw_line_heads;
extern int do_disk_light;
int alt_lighting;
float xmin, xmax, ymin, ymax, zmin, zmax, deltax, deltay, deltaz;
extern int regulus, umbra, shadow, plucker_regulae, tetrahedra, show_implicit_quadric, extrude_holes;
extern int experimental;
extern int show_twosided;
extern int writing_frames, skip_frame, first_frame, last_frame;
extern int accumulating;
extern int do_multisample;
extern int show_containment;
extern int r_density, w_density;
extern float mat_shininess;
extern float time_step, time_val;
extern float line_width;

int parse_ascii(){}
void dump_materials(){}
int termination_condition_exists(){}
char *script_record_file, *script_playback_file;

extern void REFRESH ( void );

int
main(argc,argv) 
char **argv; 
int argc;
{	
int c;
extern char *optarg;
int doreadcontours = 0;
int doreadsegments = 0;

#ifdef BUTTONFLY
    if (fork())	    /* spawn a child so buttonfly doesn't wait around */
	exit (0);
#endif

    triangulate = 0; 
    debug = 0;

  /* Parse arguments */
  if (arg_parse(argc, argv,
    "",							    "Usage plucker [options] ('plucker -' prints all options)",
    "-c",		ARG_FLAG(&doreadcontours),	    "read contours from standard input",
    "-e",		ARG_FLAG(&extrude_holes),	    "extrude holes from light vertices",
    "-i",		ARG_FLAG(&show_implicit_quadric),   "implicitize 3-line quadric",
    "-l",		ARG_FLAG(&doreadsegments),	    "read segments from standard input",
    "-p",		ARG_FLAG(&plucker_regulae),	    "compute plucker regulae",
    "-r",		ARG_FLAG(&regulus),		    "draw a regulus",
    "-t",		ARG_FLAG(&tetrahedra),		    "draw tetrahedra",
    "-u",		ARG_FLAG(&umbra),		    "compute and draw antiumbra",
    "-x",		ARG_FLAG(&experimental),	    "experimental mode",
    "-extents",		ARG_FLAG(&do_extents),	    	    "extent mode",
    "-blockers",	ARG_FLAG(&do_blockers),	    	    "blocker mode",
    "-formfactors",	ARG_FLAG(&do_ff),	    	    "formfactor mode",
    "-disklight",	ARG_FLAG(&do_disk_light),    	    "treat contour 0 as disk light source",
    "-solidangle",	ARG_FLAG(&do_solidangle),    	    "light source solid angle (ffactor) mode",
    "-s %d",		&shadow,		    	    "set shadow val to n",
    "-v %d",		&p_verbose,		    	    "set verbose debug level",
    "-containment",	ARG_FLAG(&show_containment),	    "show containment animation",
    "-shininess %f",	&mat_shininess,			    "material shininess",
    "-altlighting",	ARG_FLAG(&alt_lighting), 	    "use alternate lighting model",
    "-radius %f",	&apen_cyl_radius,		    "cylinder radius",
    "-twosided",	ARG_FLAG(&show_twosided),	    "two-sided lighting model",
    "-linewidth %f",	&line_width,	    		    "set non-lit line width",
    "-rdensity %d",	&r_density,	    		    "regulus discretization density",
    "-wdensity %d",	&w_density,	    		    "wedge discretization density",
    "-velocity %f",	&time_step,	    		    "animation time step velocity",
    "-starttime %f",	&time_val,	    		    "starting animation time [0.0..1.0]",
    "-record %S",	&script_record_file,		    "record event script to file",
    "-playback %S",	&script_playback_file,		    "playback event script from file",
    "-writeframes",	ARG_FLAG(&writing_frames),	    "write each frame as .rgb file",
    "-skipframe %d",	&skip_frame,                       "# frames to skip (fault 2)",
    "-firstframe %d",	&first_frame,                       "first frame to write (default 0)",
    "-lastframe %d",	&last_frame,                        "last frame to write (default oo)",
    "-accumulate",	ARG_FLAG(&accumulating),	    "enable accumulation buffer",
    "-multisample",	ARG_FLAG(&do_multisample),	    "multisample on VENICE h/w",
    0) < 0) exit(1);

    if (do_blockers)
	draw_line_heads = 0;
    
    if (!outputmode)	{
	init_windows (argv[0]);
	init_machine ();
	if (!alt_lighting)
	    init_lighting ();
	else
	    init_altlighting();	    
	if (mat_shininess > 0.0)
	    mshiny (mat_shininess);
	}

#if 0
    {
    RGBmode();
    doublebuffer();
    gconfig();
    lsetdepth (0, 0x7fffff);

    mmode (MPROJECTION);
    perspective((int)(340), aspect_ratio, 0.5, 10.0);
    mmode(MVIEWING);
    /* loadmatrix (idmat); */
    lookat (5.0, 5.0, 5.0, 0.0, 0.0, 0.0, 0);
    
    czclear(0x80202020, 0x007fffff);
    linewidth (2);
    draw_unit_cube_wires (0);
    draw_unit_cube_wires (1);
    draw_unit_cube_wires (2);
    swapbuffers ();

    sleep (2);
    }
#endif

    nd_set_ubailout (snapshot_all_data);

    init_grfix();
    init_geom();
    init_adt();
    init_cont();
    init_del();
    init_wing();
    init_apen();

    oldstylecontours = 1;

    if (doreadcontours)
	read_contours();    
    else if (doreadsegments)
	read_segments();

    if (script_record_file)	{
	fprintf (stderr, "recording events to file '%s'\n", script_record_file);
	tape_open (script_record_file, TAPE_RECORD);
	}
    else if (script_playback_file)	{
	fprintf (stderr, "playing back script from file '%s'\n", script_playback_file);
	tape_open (script_playback_file, TAPE_PLAYBACK);
	}

    if (shadow)    {
	read_contours();
	if ((shadow == 2 && ncontours != 3) || (shadow == 3 && ncontours < 4))
	    exit (fprintf (stderr, "must have 2->3, 3->4+ contours in input\n"));
	}
    else if (plucker_regulae)	{
	read_contours();
	}
    else if (regulus || tetrahedra) {
	if (nsegments < 0)  {
            init_segments (3);
	    nsegments = 3;
	    }
	}
    else if (ncontours == 0)	{
	if (nsegments < 0)
	    init_segments (4);
	}

    maxncontours = ncontours;

    get_bbox();
    reset_view();

    introduce();

    qenter (REDRAW, winget());
    while (1) {
	if (REACT() || DIRTY())
	    REFRESH ();
        }
}

