#include "headers.h"
#include "tape/tape.h"
#include "ff/ff.h"

#define BLOCKERS 1

#if BLOCKERS

// snarfed from pat's point3.h
typedef float Pt3Coord;
typedef struct { Pt3Coord x, y, z; } Point3;

// from hplane3.h
typedef float HPl3Coord;
typedef struct { HPl3Coord a, b, c, d; } HPlane3;

#define PARTIAL    4
#define VISIBLE    5
#define INVISIBLE  6
#define PARTIAL_HARD  7
#define VISIBLE_HARD  8

#include "plucker/blocker.h"

#endif

int pref_face = -1;
int hackcolor = 12; /* RGB_RED */
int pref_seg = -1, pref_hole = -1;
int pref_contours = -1;
int super_stabber_mode = FALSE;
int nsegments = -1;
float segments[MAX_SEGMENTS][8];	/* xyzw_1, xyzw_2 */
int dirty_data = TRUE;
int regulus = FALSE;
int plucker_regulae = FALSE;
int shadow = FALSE;
int tetrahedra = FALSE;
int umbra = FALSE;
int r_density = 25, w_density = 40;
int sphere_depth = 10;	    /* gl default is 5 */
int shaded = FALSE;
int show_twosided = FALSE;
int show_solutions = FALSE;
int show_envelope = FALSE;
int show_exploded = FALSE;
int twoway = FALSE;
int animating = FALSE;
int iomode = TAPE_QUIET;
int spinning = FALSE;
float mat_shininess = 20.0;
int show_containment = FALSE;
int printscreen = FALSE;
int suppressing_contours;
int do_multisample;  /* VENICE ONLY */
int fill_light_source = 0;
int testing_lap = 0;
int do_blockers = 0;
int do_extents = 0;
int do_ff = 0;
double form_factor = 0.0;
int do_blocker_type = 0;
int do_solidangle = 0;
int do_disk_light = 0;
int subdivideS=-1, subdivideR=-1;

float time_val = 0.0, time_step = 0.01;

int show_halfspaces = FALSE;
int show_stabbers = FALSE;
int show_antiumbra = FALSE;
int show_antipenumbra = FALSE;
int show_ieee = FALSE;
int show_ive = FALSE;
int accumulating = FALSE;
int draw_cylinders = TRUE;
int draw_line_heads = TRUE;

int show_implicit_quadric = FALSE;
int extrude_holes = FALSE;
float shadow_length = 1.414214;
float regulus_length = 1.0;
float cextrusion_distance = 1.5;
float line_width = 1.0;
int experimental = FALSE;
static int frame_num;
int writing_frames, skip_frame = 2, first_frame = 0, last_frame = 10000000;
extern int debug;
extern char *script_playback_file;
int e_j, e_k, e_m;

int ncontours, maxncontours, pref_contour = -1;
int linearize_ap = 0;

long video_colorid[16] = {
  RGB_TURQUOISE, RGB_TURTLEGREEN, RGB_MAGENTA, RGB_5,
  RGB_10, RGB_ALICEBLUE, RGB_DARKORANGE4, RGB_LAVENDER, 
  RGB_TURQUOISE, RGB_TURTLEGREEN, RGB_MAGENTA, RGB_5,
  RGB_10, RGB_ALICEBLUE, RGB_DARKORANGE4, RGB_LAVENDER 
};

/* OLD
    POINT contours[MAX_CONTOURS][MAX_LENGTH];
    int lengths[MAX_CONTOURS];
    int ceselected[MAX_CONTOURS][MAX_LENGTH];
    float cplanes[MAX_SEGMENTS][4];
*/

/* NEW: */

CONTOUR **contours = NULL;

void connect_p4();

void get_bbox();

introduce()
{
  fprintf (stderr, "ctrl-d to dirty data\n");
  fprintf (stderr, "dD to de/increase all density\n");
  fprintf (stderr, "eE to change extrusion dist/umbra scale\n\t(ctrl to toggle shadow envelope)\n");
  fprintf (stderr, "   eE also toggles portals/supporting planes\n");
  fprintf (stderr, "fF to de/increase shadow length\n");

  fprintf (stderr, "P to toggle penumbral EEE events\n");
  fprintf (stderr, "I to toggle internal VE/EEE events\n");

  fprintf (stderr, "H to toggle penumbral VE events\n");
  fprintf (stderr, "V to toggle internal VE events\n");

  fprintf (stderr, "\nL to toggle extremal stabbers\n");
  fprintf (stderr, "\nN to turn all display off\n");
  fprintf (stderr, "\n<,> to decrease/increase # of holes\n");

  fprintf (stderr, "J,K,M to cycle debug edges\n");

  fprintf (stderr, "rR to de/increase regulus length\n");
  fprintf (stderr, "S to shade/unshade\n");
  fprintf (stderr, "tT to thin/thicken line widths\n");
  fprintf (stderr, "U to umbra/penumbra\n");
  fprintf (stderr, "wW to de/increase wedge density\n");
  fprintf (stderr, "zZ to de/increase debug\n");
  fprintf (stderr, "ctrl to draw/pick baseplane contour\n");
  fprintf (stderr, "LMOUSE vtx to move vtx, edge to translate [shift to move normal]\n");
  fprintf (stderr, "  at shared edge, with LMOUSE down, press MMOUSE to jump across edge\n");
  fprintf (stderr, "alt LMOUSE edge to select/deselect\n");
  fprintf (stderr, "RMOUSE vtx to rotate\n");
  fprintf (stderr, "\n");
}

void
emit_contours ()
{
  emit_contour_list (NULL, ncontours, contours);
}

keyboard (dev, shifted)
  int dev, shifted;
{
  int cid;

  switch (dev)    {

  case ZEROKEY:
	pref_seg = 0;
	break;

  case ONEKEY:
	pref_seg = 1;
	break;

  case TWOKEY:
	pref_seg = 2;
	break;

  case THREEKEY:
	pref_seg = 3;
	break;

  case FOURKEY :
	pref_seg = 4;
	break;

	// num pad
  case   PAD7:
	if (shifted)
	  subdivideS = C_NW;
	else
	  subdivideR = C_NW;
	break;

  case   PAD1:
	if (shifted)
	  subdivideS = C_SW;
	else
	  subdivideR = C_SW;
	break;

  case   PAD3:
	if (shifted)
	  subdivideS = C_SE;
	else
	  subdivideR = C_SE;
	break;

  case   PAD9:
	if (shifted)
	  subdivideS = C_NE;
	else
	  subdivideR = C_NE;
	break;

  case SEVENKEY :
	suppressing_contours = 1 - suppressing_contours;
	break;

  case PRINTSCREENKEY :
	printscreen = 1;
	break;

  case AKEY:
	if (!shifted)
	  accumulating = 1 - accumulating;
	else
	  animating = 1-animating;
	break;

  case BKEY:
	if (control)	{
	  if (!shifted)   {
		get_bbox();
		reset_view();
	  }
	  else    {
		do_blocker_type = !do_blocker_type;
	  }
	}
	else if (do_blockers)	{
	  if (shifted)
		do_blockers++;
	  else if (do_blockers > 1)
		do_blockers--;
	}
	break;

  case CKEY :
	if (shifted)
	  apen_cyl_radius *= 1.05;
	else
	  apen_cyl_radius /= 1.05;
	fprintf (stderr, "apen cyl radius now %8.5f\n", apen_cyl_radius);
	break;

  case DKEY:
	if (control)	{
	  fprintf (stderr, "dirtying data\n");    
	  dirty_data = 1;
	  break;
	}
		
	if (ncontours > 1 || nsegments > 1)  {    
	  if (shifted)	{
		w_density += 2;
		r_density += 2;
	  }
	  else	{
		w_density -= 2;
		r_density -= 2;
	  }

	  if (w_density < 0)
		w_density = 0;

	  if (r_density < 0)
		r_density = 0;

	  fprintf (stderr, "w, r density now %d %d\n", w_density, r_density);
	}
	else	{
	  if (shifted)    {
		if (sphere_depth < SPH_MAXDEPTH)
		  sphere_depth++;
	  } 
	  else    {
		if (sphere_depth > 1)
		  sphere_depth--;
	  }
	  fprintf (stderr, "sphere depth now %d\n", sphere_depth);
	}
	break;

  case EKEY:
	if (!control)
	  show_exploded = 1 - show_exploded;
	else	{
	  /* show_envelope = 1-show_envelope; */
	  cextrusion_distance *= (shifted) ? 1.1 : 0.9;
	}
	break;

  case FKEY:  
	if (control)
	  fill_light_source = 1- fill_light_source;
	else if (shifted)
	  shadow_length *= 1.1;
	else
	  shadow_length /= 1.1;
	break;

  case GKEY:
	graph_normals_only = !graph_normals_only;
	dirty_data = TRUE;
	break;

  case HKEY:
	if (!control)
	  show_halfspaces = 1-show_halfspaces;
	else	{
	  if (!shifted)
		r_density -= 5;
	  else
		r_density += 5;
	  fprintf (stderr, "w, r density now %d %d\n", w_density, r_density);
	  dirty_data = TRUE;
	}
	break;

  case IKEY:
	show_ieee = 1 - show_ieee;
	show_ive = 1 - show_ive;
	break;

  case JKEY:
	if (shifted)
	  hackcolor++;
	else
	  hackcolor--;
	fprintf (stderr, "hackcolor now %d\n", hackcolor);
	break;
	
  case KKEY:
	if (shifted)
	  pref_hole++;
	else
	  pref_hole--;    
	break;

  case NKEY:  
	if (!control)	{
	  if (shifted)
		pref_seg++;
	  else 
		pref_seg--;
	  fprintf (stderr, "pref stabber now %d\n", pref_seg);
	}
	else	
	  show_ieee = show_ive = show_stabbers = show_antipenumbra = show_antiumbra = show_halfspaces = 0;
	break;

  case LKEY:
	if (!shifted)
	  show_stabbers = 1 - show_stabbers;
	else if (!control) {
	  linearize_ap = (linearize_ap + 1) % 6;
	  dirty_data = 1;
	  fprintf (stderr, "linearize now %d: ", linearize_ap);
	  switch (linearize_ap)   {
	  case 0 : fprintf (stderr, "quadratic antipen\n");
		break;
	  case 1 : fprintf (stderr, "linearize antipen\n");
		break;
	  case 2 : fprintf (stderr, "antiumbra (linear)\n");
		break;
	  case 3 : fprintf (stderr, "antiumbra + linear apen\n");
		break;
	  case 4 : fprintf (stderr, "quad + linear apen\n");
		break;
	  case 5 : fprintf (stderr, "quad + linear apen + antiumbra\n");
		break;
	  }
	}
	else
	  draw_line_heads = 1-draw_line_heads;
	break;

  case OKEY:
	if (!shifted)
	  dirty_data = p_armed = TRUE;
	else {
	  p_randomize = !p_randomize;
	  fprintf (stderr, "randomizing now %d\n", p_randomize);
	}
	break;

  case PKEY:
	if (!shifted)
	  show_antipenumbra = 1 - show_antipenumbra;
	else	{
	  for (cid = 0; cid < ncontours; cid++)
		planarize_contour (contours[cid]);
	}
	break;

  case QKEY:
	testing_lap = 1-testing_lap;
	break;

  case RKEY:  
	if (shifted)
	  regulus_length *= 1.1;
	else
	  regulus_length /= 1.1;
	break;

  case SKEY:
	if (!control)
	  shaded = !shaded;
	else	{
	  mat_shininess *= (shifted) ? 1.05 : 0.95;
	  mshiny (mat_shininess);
	  fprintf (stderr, "mat shininess now %f\n", mat_shininess);
	}
	break;

  case TKEY:
	if (!control)	{   
	  twoway = 1-twoway;
	  dirty_data = TRUE;
	}
	else	{
	  if (shifted)
		line_width *= 1.1;
	  else
		line_width /= 1.1;
	  if (line_width < 1.0)
		line_width = 1.0;
	  fprintf (stderr, "gl line width now %f\n", line_width);
	}
	break;

  case UKEY:
	show_antiumbra = 1 - show_antiumbra;
	umbra = 1 - umbra;
	break;

  case VKEY:
	if (control)	
	  show_ive = 1 - show_ive;
	else	{
	  if (shifted)
		time_step *= 1.05;
	  else
		time_step /= 1.05;
	  fprintf (stderr, "time step now %8.4f\n", time_step);
	}
	break;

  case WKEY:
	if (shifted)
	  w_density += 2;
	else
	  w_density -= 2;
	if (w_density < 1)
	  w_density = 1;
	break;

  case XKEY:
	spinning = 1-spinning;
	break;

  case ZKEY:
	if (!control)	{
	  debug += (shifted) ? 1 : -1;
	  fprintf (stderr, "debug now %d\n", debug);
	}
	else	{
	  extern int projectz;

	  projectz = 1-projectz;
	}
	break;

  case COMMAKEY :
	if (ncontours > 1)  {
	  ncontours--;
	  dirty_data = TRUE;
	}
	break;

  case PERIODKEY :
	if (ncontours < maxncontours)	{
	  ncontours++;
	  dirty_data = TRUE;
	}
	break;

  default :
	fprintf (stderr, "unknown event %d\n", dev); 
	introduce();
	break;
  }
}		

/* uses global draw_line_heads */
void
render_oriented_line (
					  POINT *a,
					  POINT *mid,
					  POINT *b,
					  int draw_cylinder,
					  float cylinder_scale,
					  long bodycol,
					  long headcol)
{
  if (!draw_line_heads)
	mid = b;

  if (!draw_cylinders) {
	CPACK (bodycol);
	bgnline ();
	v3f ((float *)a);
	v3f ((float *)mid);
	endline ();

	if (!draw_line_heads)
	  return;

	CPACK (headcol);
	bgnline ();
	v3f ((float *)mid);
	v3f ((float *)b);
	endline ();
  }
  else    {
	MPACK (bodycol);
	capped_cylinder ((POINT *)a, (POINT *)mid, apen_cyl_radius * cylinder_scale, 16);

	if (!draw_line_heads)
	  return;

	MPACK (headcol);
#if 0
	capped_cylinder ((POINT *)mid, (POINT *)b, apen_cyl_radius * cylinder_scale, 16);
#else
	arrowhead ((POINT *)mid, (POINT *)b, apen_cyl_radius * cylinder_scale, 16);
#endif
  }
}

static int view_reset = FALSE;

void
read_segments ()
{
  int k;

  k = nsegments = 0;
  while (1)	{
    if (scanf ("%f %f %f\n", &segments[k][0], &segments[k][1], &segments[k][2]) != 3)
	  break;
	if (scanf ("%f %f %f\n", &segments[k][4], &segments[k][5], &segments[k][6]) != 3)
	  exit (fprintf (stderr, "expecting x, y, z floats for seg %d endpt 0\n", k));
	segments[k][3] = segments[k][7] = 1.0;
	fprintf (stderr, "got line <%g %g %g %g> <%g %g %g %g>\n",
			 segments[k][0], segments[k][1],segments[k][2], segments[k][3],
			 segments[k][4], segments[k][5], segments[k][6], segments[k][7]);
	k++;
	nsegments++;

    /* XXX should extend and/or clip segment so that its endpoints lie on unit cube */
	/* problem -- if it is disjoint from unit cube, should we discard it and exit? */

  }
  fprintf (stderr, "read_segments(): found %d input segments\n", nsegments);

}

void
get_bbox()
{
  int k, n;

  if (ncontours == 0)	{
	xmin = ymin = zmin = -1.0;
	xmax = ymax = zmax =  1.0;
	return;
  }

  xmin = ymin = zmin =  1.0E10;
  xmax = ymax = zmax = -1.0E10;

  for (k = 0; k < ncontours; k++)	{
	for (n = 0; n < CLENGTH(contours[k]); n++)    {
	  xmin = MIN (xmin, CX(contours[k], n));
	  ymin = MIN (ymin, CY(contours[k], n));
	  zmin = MIN (ymin, CZ(contours[k], n));

	  xmax = MAX (xmax, CX(contours[k], n));
	  ymax = MAX (ymax, CY(contours[k], n));
	  zmax = MAX (zmax, CZ(contours[k], n));
	}
  }
}

void
read_contours()
{
  int k, n, clength, nscanned;
  CONTOUR *c;
  long ccolor;
  static char line[80];

  scanf ("%d\n", &ncontours);

  if (ncontours <= 0)
	exit (fprintf (stderr, "expecting # of contours\n"));

  contours = (CONTOUR **) malloc ((ncontours+1) * sizeof (CONTOUR *));
  fprintf (stderr, "reading %d contours... ", ncontours);

  for (k = 0; k < ncontours; k++)	{

	if (gets (line) == NULL)
	  exit (fprintf (stderr, "couldn't read # of points line for contour %d\n", k));

	nscanned = sscanf (line, "%d 0x%x\n", &clength, &ccolor);
	if (nscanned < 1)
	  exit (fprintf (stderr, "expecting # of points > 0 for contour %d\n", k));

	contours[k] = c = create_dummy_colored_contour (clength);
	/*
	  if (verbose)
	  fprintf (stderr, "reading contour %d (%d points)\n", k, clength);
	  */
	for (n = 0; n < clength; n++)    {
	  if (scanf ("%f %f %f\n", &CX(c, n), &CY(c, n), &CZ(c, n)) != 3)
		exit (fprintf (stderr, "expecting x, y, z floats for contour %d pt %d\n", k, n));
	}

	find_contour_attributes (c);
	if (nscanned == 2)
	  imbue_contour_color (c, ccolor);
  }

  fprintf (stderr, "done.\n");
  contours[ncontours] = NULL;

  xmin = ymin = zmin = -1.0;
  xmax = ymax = zmax =  1.0;
}

CONTOUR *
dopple_contour(CONTOUR *c, float edist)
{
  int k;
  float dx, dy, dz;
  CONTOUR *d;

  d = copy_contour (c);

  dx = edist * CP_A(c);
  dy = edist * CP_B(c);
  dz = edist * CP_C(c);

  CP_D(d) = CP_D(c) - edist;

  for (k = 0; k < CLENGTH(c); k++)	{
	CX(d, k) = CX(c, k) + dx;
	CY(d, k) = CY(c, k) + dy;
	CZ(d, k) = CZ(c, k) + dz;
  }

  return d;
}

/* returns vertexid if picking */
int
draw_contour (
			  CONTOUR *cc,
			  int picking_vertices,
			  int picking_edges,
			  int docylinders,
			  long col, long prefcol)
{
  int pickid, k, len;
  POINT *c, *vA, *vB, _vP, *vP= &_vP, _vQ, *vQ = &_vQ;
  char str[80];

  if (picking_vertices || picking_edges)
	setup_picking();
  else
	linewidthf (2.5);

  if (docylinders)	{
	push_lighting(1);
	MPACK (col);
  }
  else
	CPACK (col);

  len = CLENGTH(cc);
  c = CPOINTS(cc);

  if (picking_vertices)    
	for (k = 0; k < len; k++)       {
	  vA = &c[(k + len - 1) % len];
	  vB = &c[(k + 1) % len];

	  vlerp ((VECTOR *)vP, (VECTOR *)vA, (VECTOR *)&c[k], 0.9);
	  vlerp ((VECTOR *)vQ, (VECTOR *)&c[k], (VECTOR *)vB, 0.1);
            
	  loadname (k);
	  bgnline();
	  v3f ((float *)vP);
	  v3f ((float *)&c[k]);
	  v3f ((float *)vQ);
	  endline();
	}
  else    {
	if (!picking_edges)  {	    /* non-picking case */
	  for (k = 0; k < len; k++)	{   /* < len for closed poly; < len - 1 for open poly  */
		vA = &c[k];
		vB = &c[(k+1) % len];	/* (k+1) % len for closed poly *//* (k+1) for open poly */
		plerp (vP, vA, vB, 0.8);
		render_oriented_line (vA, vP, vB, docylinders, 1.0, col, prefcol);

		if (len == 2)		/* closed poly, 2 vertices */
		  break;
	  }
	}
	else    {	/* picking edges */
	  for (k = 0; k < len; k++)       {   /* < len for closed poly; < len - 1 for open poly  */
		loadname (k);
		vA = &c[k];
		vB = &c[(k+1) % len];
		bgnline();
		v3f ((float *)vA);
		v3f ((float *)vB);
		endline();
	  }
	}
  }

  linewidthf (1.0);

  if (docylinders)
	pop_lighting();

  return ((picking_vertices || picking_edges) ? first_pick() : -1);
}

int
draw_filled_contour (
					 CONTOUR *cc,
					 long col)
{
  int k, len;
  POINT *c;

  if (lighting)   {
	MPACK (col);
	n3f ((float *)CNORMAL(cc));
  }
  else 
	CPACK (col);

  len = CLENGTH(cc);
  c = CPOINTS(cc);
  bgnpolygon();
  for (k = 0; k < len; k++)
	v3f ((float *)&c[k]);
  endpolygon();
}

/* segs are constrained to have endpoints on faces of 1,1,1 cube */
init_segments (int nsegs)
{
  int k;
  int f;

  nsegments = nsegs;
  for (k = 0; k < nsegs; k++) {
	f = rand () % 3;

	segments[k][f] = -1.0;
	segments[k][(f + 1) % 3] = randr (-1.0, 1.0);
	segments[k][(f + 2) % 3] = randr (-1.0, 1.0);
	segments[k][3] = 1.0;

	segments[k][4 + f] =  1.0;
	segments[k][4 + (f + 1) % 3] = randr (-1.0, 1.0);
	segments[k][4 + (f + 2) % 3] = randr (-1.0, 1.0);
	segments[k][7] = 1.0;
  }

  xmin = ymin = zmin = -1.0;
  xmax = ymax = zmax =  1.0;
}

long scol[4] = {RGB_RED, RGB_GREEN, RGB_BLUE, RGB_ORANGE};

float vNNN[3] = {-1.0, -1.0, -1.0};
float vNNP[3] = {-1.0, -1.0,  1.0};
float vNPN[3] = {-1.0,  1.0, -1.0};
float vNPP[3] = {-1.0,  1.0,  1.0};
float vPNN[3] = { 1.0, -1.0, -1.0};
float vPNP[3] = { 1.0, -1.0,  1.0};
float vPPN[3] = { 1.0,  1.0, -1.0};
float vPPP[3] = { 1.0,  1.0,  1.0};

float nPX[3] = { 1.0,  0.0,  0.0 };
float nNX[3] = {-1.0,  0.0,  0.0 };
float nPY[3] = { 0.0,  1.0,  0.0 };
float nNY[3] = { 0.0, -1.0,  0.0 };
float nPZ[3] = { 0.0,  0.0,  1.0 };
float nNZ[3] = { 0.0,  0.0, -1.0 };

void
draw_unit_cube_faces (int pface)
{
  push_lighting(1);
  blending_on();

  CPACK (RGB_MAGENTA);
    	
  if (pface == 0)	{
	bgnpolygon();
	n3f ((float *)nPX);
	v3f ((float *)vPNN);
	v3f ((float *)vPPN);
	v3f ((float *)vPPP);
	v3f ((float *)vPNP);
	endpolygon();

	bgnpolygon();
	n3f ((float *)nNX);
	v3f ((float *)vNNN);
	v3f ((float *)vNPN);
	v3f ((float *)vNPP);
	v3f ((float *)vNNP);
	endpolygon();
  }
  else if (pface == 1)	{
	bgnpolygon();
	n3f ((float *)nPY);
	v3f ((float *)vPPN);
	v3f ((float *)vNPN);
	v3f ((float *)vNPP);
	v3f ((float *)vPPP);
	endpolygon();

	bgnpolygon();
	n3f ((float *)nNY);
	v3f ((float *)vPNN);
	v3f ((float *)vNNN);
	v3f ((float *)vNNP);
	v3f ((float *)vPNP);
	endpolygon();
  }
  else if (pface == 2)	{
	bgnpolygon();
	n3f ((float *)nNZ);
	v3f ((float *)vNNN);
	v3f ((float *)vPNN);
	v3f ((float *)vPPN);
	v3f ((float *)vNPN);
	endpolygon();

	bgnpolygon();
	n3f ((float *)nPZ);
	v3f ((float *)vNNP);
	v3f ((float *)vPNP);
	v3f ((float *)vPPP);
	v3f ((float *)vNPP);
	endpolygon();
  }

  blending_off();
  pop_lighting();
}

void
draw_unit_cube_wires (int pface)
{
  /* CPACK (RGB_MAGENTA); */
  cpack (0x0000ff00);
    	
  if (pface == 0)	{
	bgnclosedline();
	v3f ((float *)vPNN);
	v3f ((float *)vPPN);
	v3f ((float *)vPPP);
	v3f ((float *)vPNP);
	endclosedline();

	bgnclosedline();
	v3f ((float *)vNNN);
	v3f ((float *)vNPN);
	v3f ((float *)vNPP);
	v3f ((float *)vNNP);
	endclosedline();
  }
  else if (pface == 1)	{
	bgnclosedline();
	v3f ((float *)vPPN);
	v3f ((float *)vNPN);
	v3f ((float *)vNPP);
	v3f ((float *)vPPP);
	endclosedline();

	bgnclosedline();
	v3f ((float *)vPNN);
	v3f ((float *)vNNN);
	v3f ((float *)vNNP);
	v3f ((float *)vPNP);
	endclosedline();
  }
  else if (pface == 2)	{
	bgnclosedline();
	v3f ((float *)vNNN);
	v3f ((float *)vPNN);
	v3f ((float *)vPPN);
	v3f ((float *)vNPN);
	endclosedline();

	bgnclosedline();
	v3f ((float *)vNNP);
	v3f ((float *)vPNP);
	v3f ((float *)vPPP);
	v3f ((float *)vNPP);
	endclosedline();
  }
}

void
connect4 (center, A, B, C, D, doit)
  float *center;
float *A, *B, *C, *D;
int doit;
{
  bgnclosedline();
  v3f ((float *)A);
  v3f ((float *)B);
  v3f ((float *)C);
  v3f ((float *)D);
  endclosedline();

  if (!doit)
	return;

  setlinestyle (DASHED_LINE);

  bgnline ();
  v3f ((float *)center);
  v3f ((float *)A);
  endline ();

  bgnline ();
  v3f ((float *)center);
  v3f ((float *)B);
  endline ();

  bgnline ();
  v3f ((float *)center);
  v3f ((float *)C);
  endline ();

  bgnline ();
  v3f ((float *)center);
  v3f ((float *)D);
  endline ();

  setlinestyle (SOLID_LINE);
}

static int disabled[6];

void
draw_unit_cube ()
{
  float *Ap, *Bp, *Cp, *Dp, *An, *Bn, *Cn, *Dn;
  int sid, off;
  int face1, face2;
  int k;

  disabled[0] = disabled[1] = disabled[2] = disabled[3] = disabled[4] = disabled[5] = FALSE;

  sid = pref_seg / 2;
  off = (pref_seg & 1) << 2;

  if (segments[sid][off] == 1.0)  {	/* +x */
	disabled[0] = TRUE;
	CPACK (pref_face == 0 ? RGB_RED : RGB_WHITE);
	connect4 (&segments[sid][off], vPNN, vPPN, vPPP, vPNP, pref_face == 0);
  }

  if (segments[sid][off] == -1.0) {   /* -x */
	disabled[1] = TRUE;
	CPACK (pref_face == 0 ? RGB_RED : RGB_WHITE);
	connect4 (&segments[sid][off], vNNN, vNPN, vNPP, vNNP, pref_face == 0);
  }

  if (segments[sid][off + 1] ==  1.0)	{   /* +y */
	disabled[2] = TRUE;
	CPACK (pref_face == 1 ? RGB_RED : RGB_WHITE);
	connect4 (&segments[sid][off], vPPN, vNPN, vNPP, vPPP, pref_face == 1);
  }

  if (segments[sid][off + 1] ==  -1.0)    {   /* -y */
	disabled[3] = TRUE;
	CPACK (pref_face == 1 ? RGB_RED : RGB_WHITE);
	connect4 (&segments[sid][off], vPNN, vNNN, vNNP, vPNP, pref_face == 1);
  }

  if (segments[sid][off + 2] == 1.0)	{   /* +z */
	disabled[4] = TRUE;
	CPACK (pref_face == 2 ? RGB_RED : RGB_WHITE);
	connect4 (&segments[sid][off], vNNP, vPNP, vPPP, vNPP, pref_face == 2);
  }

  if (segments[sid][off + 2] ==  -1.0)    {      /* -z */
	disabled[5] = TRUE;
	CPACK (pref_face == 2 ? RGB_RED : RGB_WHITE);
	connect4 (&segments[sid][off], vNNN, vPNN, vPPN, vNPN, pref_face == 2);
  }

  off = off ? off - 4 : off + 4;

  if (!disabled[0] && segments[sid][off] == 1.0)  {	/* +x */
	CPACK (RGB_BLUE);
	connect4 (&segments[sid][off], vPNN, vPPN, vPPP, vPNP, TRUE);
  }

  if (!disabled[1] && segments[sid][off] == -1.0) {   /* -x */
	CPACK (RGB_BLUE);
	connect4 (&segments[sid][off], vNNN, vNPN, vNPP, vNNP, TRUE);
  }

  if (!disabled[2] && segments[sid][off + 1] ==  1.0)	{   /* +y */
	CPACK (RGB_BLUE);
	connect4 (&segments[sid][off], vPPN, vNPN, vNPP, vPPP, TRUE);
  }

  if (!disabled[3] && segments[sid][off + 1] ==  -1.0)    {   /* -y */
	CPACK (RGB_BLUE);
	connect4 (&segments[sid][off], vPNN, vNNN, vNNP, vPNP, TRUE);
  }

  if (!disabled[4] && segments[sid][off + 2] == 1.0)	{   /* +z */
	CPACK (RGB_BLUE);
	connect4 (&segments[sid][off], vNNP, vPNP, vPPP, vNPP, TRUE);
  }

  if (!disabled[5] && segments[sid][off + 2] ==  -1.0)    {      /* -z */
	CPACK (RGB_BLUE);
	connect4 (&segments[sid][off], vNNN, vPNN, vPPN, vNPN, TRUE);
  }
}

/* construct implicit equation of quadric */
void
draw_implicit_regulus (
					   POINT *p, POINT *q,
					   POINT *r, POINT *s,
					   POINT *t, POINT *u)
{
  float Ax,Bx,Cx,Dx,
	Ay,By,Cy,Dy,
	Az,Bz,Cz,Dz,
	Aw,Bw,Cw,Dw;
  float A,B,C,D,E,F,G,H,I,J;
  MATRIX Q;
  int i,j,k;
  POINT P;
  float dx,dy,dz;
  float dist, margin;
  unsigned short red, green;
  static POINT plist[8192];
  static unsigned long clist[8192];
  static int npts;

  if (!dirty_data)	{   /* draw stored points and return */
	pntsize (3);
	for (k = 0; k < npts; k++)  {
	  cpack (clist[k]);
	  pnt (plist[k].px, plist[k].py, plist[k].pz);
	}
	pntsize (1);
	return;
  }

  /* otherwise recompute points */

  /* compute minors of implicit equation */
  Ax = det33 (p->py, p->pz, 1.0, q->py, q->pz, 1.0, r->py, r->pz, 1.0);
  Bx = det33 (s->py, s->pz, 1.0, t->py, t->pz, 1.0, u->py, u->pz, 1.0);
  Cx = det33 (p->py, p->pz, 1.0, q->py, q->pz, 1.0, s->py, s->pz, 1.0);
  Dx = det33 (r->py, r->pz, 1.0, t->py, t->pz, 1.0, u->py, u->pz, 1.0);

  Ay = det33 (p->px, p->pz, 1.0, q->px, q->pz, 1.0, r->px, r->pz, 1.0);
  By = det33 (s->px, s->pz, 1.0, t->px, t->pz, 1.0, u->px, u->pz, 1.0);
  Cy = det33 (p->px, p->pz, 1.0, q->px, q->pz, 1.0, s->px, s->pz, 1.0);
  Dy = det33 (r->px, r->pz, 1.0, t->px, t->pz, 1.0, u->px, u->pz, 1.0);

  Az = det33 (p->px, p->py, 1.0, q->px, q->py, 1.0, r->px, r->py, 1.0);
  Bz = det33 (s->px, s->py, 1.0, t->px, t->py, 1.0, u->px, u->py, 1.0);
  Cz = det33 (p->px, p->py, 1.0, q->px, q->py, 1.0, s->px, s->py, 1.0);
  Dz = det33 (r->px, r->py, 1.0, t->px, t->py, 1.0, u->px, u->py, 1.0);

  Aw = det33 (p->px, p->py, p->pz, q->px, q->py, q->pz, r->px, r->py, r->pz);
  Bw = det33 (s->px, s->py, s->pz, t->px, t->py, t->pz, u->px, u->py, u->pz);
  Cw = det33 (p->px, p->py, p->pz, q->px, q->py, q->pz, s->px, s->py, s->pz);
  Dw = det33 (r->px, r->py, r->pz, t->px, t->py, t->pz, u->px, u->py, u->pz);

  /* compute coefficients of 4x4 quadric matrix */

  /* quadratic */
  A = Ax * Bx - Cx * Dx;
  E = Ay * By - Cy * Dy;
  H = Az * Bz - Cz * Dz;

  /* bilinear */
  B =  0.5 * (Cx * Dy + Cy * Dx - Ax * By - Ay * Bx);
  C = -0.5 * (Cx * Dz + Cz * Dx - Ax * Bz - Az * Bx);
  F =  0.5 * (Cy * Dz + Cz * Dy - Ay * Bz - Az * By);

  /* linear */
  D = -0.5 * (Ax * Bw + Aw * Bx - Cx * Dw - Cw * Dx);
  G =  0.5 * (Ay * Bw + Aw * By - Cy * Dw - Cw * Dy);
  I = -0.5 * (Az * Bw + Aw * Bz - Cz * Dw - Cw * Dz);

  /* constant */
  J = Aw * Bw - Cw * Dw;

  /* build symmetric quadric matrix */
  loadquadric (Q, 
			   A, B, C, D,
			   B, E, F, G,
			   C, F, H, I,
			   D, G, I, J);

  /* make it unitary */
  make_unitary (Q);

  /* now point sample inside bbox of three lines */
  /* use global density, global bbox */
  dx = (xmax - xmin) / r_density;
  dy = (ymax - ymin) / r_density;
  dz = (zmax - zmin) / r_density;

  pntsize (3);
  npts = 0;
  margin = 0.02;
  for (i = 0; i <= r_density; i++)	{
	P.px = xmin + i * dx;
	for (j = 0; j <= r_density; j++)    {
	  P.py = ymin + j * dy;
	  for (k = 0; k <= r_density; k++)	{
		if (npts == 8192)
		  return;
		P.pz = zmin + k * dz;
		dist = pMpT (&P, Q);
		if (dist >= -margin && dist <= margin)    {
		  dist = (margin + dist) * 0.5 / margin;
		  red = 255 * dist;
		  green = 255 - red;
		  cpack (clist[npts] = (unsigned long)(green << 16 | red));
		  pnt (P.px, P.py, P.pz);
		  plist[npts++] = P;
		}
	  }
	}
  }
  pntsize (1);
}

/* nsegments == 2 */
void
draw_twoline_distance (
					   long planec, long stickc, long pointc)
{
  int k;
  float L[2][6];
  float dx, dy, lldist;	/* line-line distance */
  POINT P, Q;
  static POINT Org = {0.0, 0.0, 0.0};

  for (k = 0; k < 2; k++) {
	p_construct (segments[k][0], segments[k][1], segments[k][2], segments[k][3],
	             segments[k][4], segments[k][5], segments[k][6], segments[k][7], L[k]);
	p_normalize (L[k], 6);
  }

  /* take the plucker dot product */
  lldist = -2.0 * p_product (L[0], L[1]);

  /* draw the base plane at Z  =  0 */
  dx = xmax - xmin;
  dy = ymax - ymin;

  pushmatrix ();
  /* place origin at the data centroid, moved in x one bbox unit */
  translate (AVG(xmin,xmax) + dx * 1.25, 0.0, AVG(zmin, zmax));

  /* erect altitude to point */
  MPACK (stickc);
  P.px = Q.px = 0.0;
  P.py = Q.py = 0.0;
  P.pz = 0.0;
  Q.pz = lldist;
  uncapped_cylinder (&P, &Q, apen_cyl_radius * 0.75, 16);

  /* draw the point at lldist above baseplane */
  MPACK (pointc);
  draw_glsphere (0.0, 0.0, lldist, apen_cyl_radius * 1.5);

  /* draw the baseplane */
  dx *= 0.5;
  dy *= 0.5;
  MPACK (planec);
  rounded_box (&Org, &PXHAT, &PYHAT, -dx, -dy, dx, dy, apen_cyl_radius * 0.25, 16);

  popmatrix ();
}

void
draw_containment_lines (void)
{
  POINT *a, *b, *c, *d, AC, BC, AD, BD, AB, CD, ABCP, ABCM;
  VECTOR Dvec;
  float t;

  draw_segments (FALSE, 0, 0, RGB_WHITE, RGB_WHITE);
  draw_segments (FALSE, 1, 2, RGB_GREEN, RGB_GREEN);

  a = (POINT *)&segments[0][0];
  b = (POINT *)&segments[0][4];
  c = (POINT *)&segments[1][0];
  d = (POINT *)&segments[2][0];

  MPACK (RGB_RED);
  draw_sphere (a, apen_cyl_radius * 1.5);
  draw_sphere (b, apen_cyl_radius * 1.5);
  draw_sphere (c, apen_cyl_radius * 1.5);
  draw_sphere (d, apen_cyl_radius * 1.5);

  pplusv (&AC, c, vdiffp (&Dvec, c, a));
  pplusv (&BC, c, vdiffp (&Dvec, c, b));
  pplusv (&AD, d, vdiffp (&Dvec, d, a));
  pplusv (&BD, d, vdiffp (&Dvec, d, b));

  MPACK (RGB_WHITE);
  /* draw antiumbra */
  face4pn (c, d, &BD, &AC, &PZHAT);

  /* draw antipenumbrae */
  MPACK (RGB_GREY10);
  face3pn (c, &AC, &BC, &PZHAT);
  face3pn (d, &AD, &BD, &PZHAT);

  MPACK (RGB_WHITE);
  capped_cylinder (a, &AC, apen_cyl_radius * 0.5, 16);
  capped_cylinder (b, &BD, apen_cyl_radius * 0.5, 16);

  MPACK (RGB_YELLOW);
  capped_cylinder (a, &AD, apen_cyl_radius * 0.5, 16);
  capped_cylinder (b, &BC, apen_cyl_radius * 0.5, 16);

  MPACK (RGB_CYAN);
  /* lerp a contained line; pivot on A or C */
  if (animating && pref_seg == 0)  {
	if (time_step >= 0.0)   {
	  /* maps 0.0 to 0.0, 0.5 to 0.5, 1.0 to 0.0 */
	  t = 0.5 - fabsf (time_val - 0.5);
	  plerp (&AB, a, b, t);
	  pplustv (&ABCP, c, 1.25, vdiffp (&Dvec, c, &AB));
	  pminustv (&ABCM, &AB, 1.25, vdiffp (&Dvec, c, &AB));
	  capped_cylinder (&ABCM, &ABCP, apen_cyl_radius * 0.75, 16);
	}
	else    {
	  /* maps 0.0 to 0.0, 0.5 to 0.5, 1.0 to 0.0 */
	  t = 0.5 - fabsf (time_val - 0.5);
	  plerp (&CD, c, d, t);
	  pplustv (&ABCP, a, 1.25, vdiffp (&Dvec, a, &CD));
	  pminustv (&ABCM, &CD, 1.25, vdiffp (&Dvec, a, &CD));
	  capped_cylinder (&ABCM, &ABCP, apen_cyl_radius * 0.75, 16);
	}
	MPACK (RGB_MAGENTA);
	capped_cylinder (a, c, apen_cyl_radius * 0.6, 16);
  }

  /* lerp an uncontained line; pivot on B or C */
  if (animating && pref_seg == 1)  {
	if (time_step >= 0.0)   {
	  /* maps 0.0 to 0.0, 0.5 to 0.5, 1.0 to 0.0 */
	  t = 0.5 - fabsf (time_val - 0.5);
	  plerp (&AB, b, a, t);
	  pplustv (&ABCP, c, 1.25, vdiffp (&Dvec, c, &AB));
	  pminustv (&ABCM, &AB, 1.25, vdiffp (&Dvec, c, &AB));
	  capped_cylinder (&ABCM, &ABCP, apen_cyl_radius * 0.75, 16);
	}
	else    {
	  /* maps 0.0 to 0.0, 0.5 to 0.5, 1.0 to 0.0 */
	  t = 0.5 - fabsf (time_val - 0.5);
	  plerp (&CD, d, c, 1.0 - t);
	  pplustv (&ABCP, b, 1.25, vdiffp (&Dvec, b, &CD));
	  pminustv (&ABCM, &CD, 1.25, vdiffp (&Dvec, b, &CD));
	  capped_cylinder (&ABCM, &ABCP, apen_cyl_radius * 0.75, 16);
	}
	MPACK (RGB_MAGENTA);
	capped_cylinder (b, c, apen_cyl_radius * 0.6, 16);
  }
}

int
draw_segments (
			   int picking,
			   int k0, int k1,
			   long col, long prefcol)
{
  int k;
  float mid[4];

  if (picking)
	setup_picking();

  for (k = k0; k <= k1; k++) {
	if (!picking)	{
	  mid[0] = (4.0 * segments[k][4] + segments[k][0]) / 5.0;
	  mid[1] = (4.0 * segments[k][5] + segments[k][1]) / 5.0;
	  mid[2] = (4.0 * segments[k][6] + segments[k][2]) / 5.0;

	  render_oriented_line ((POINT *)&segments[k][0], (POINT *)mid, (POINT *)&segments[k][4], 
							draw_cylinders, 1.0, k == pref_seg && !show_containment ? RGB_MAGENTA : col, prefcol);
	}
	else	{
	  CPACK (RGB_ABYSS);

	  loadname (k + k);
	  bgnline ();
	  v3f ((float *)&segments[k][0]);
	  mid[0] = (3.0 * segments[k][0] + segments[k][4]) / 4.0;
	  mid[1] = (3.0 * segments[k][1] + segments[k][5]) / 4.0;
	  mid[2] = (3.0 * segments[k][2] + segments[k][6]) / 4.0;
	  v3f ((float *)mid);
	  endline ();

	  loadname (k + k + 1);
	  bgnline ();
	  mid[0] = (segments[k][0] + 3.0 * segments[k][4]) / 4.0;
	  mid[1] = (segments[k][1] + 3.0 * segments[k][5]) / 4.0;
	  mid[2] = (segments[k][2] + 3.0 * segments[k][6]) / 4.0;
	  v3f ((float *)mid);
	  v3f ((float *)&segments[k][4]);
	  endline ();
	}
  }

  return (picking ? first_pick() : -1);
}

float dist_list[128];
float stab_list[128][4];

void
connect_planes (
				float L[6],
				CONTOUR *c[],
				int nplanes,
				long codd, long ceven)
{
  int j, k, max;
  float s,t, hat;

  for (k = 0; k < nplanes; k++)   {
	p_intersect (L, (float *)CPLANE(c[k]), stab_list[k]);
	if (fabs(stab_list[k][3]) <= VEPS)
	  return;
	p_project (stab_list[k], 4);
	if (k == 0)
	  dist_list[k] = 0.0;	
	else if (k == 1)    {
	  dist_list[k] = 1.0;
	  max = 0;
	  if (fabs(stab_list[1][1] - stab_list[0][1]) > fabs(stab_list[1][max] - stab_list[0][max]))
		max = 1;
	  if (fabs(stab_list[1][2] - stab_list[0][2]) > fabs(stab_list[1][max] - stab_list[0][max]))
		max = 2;
	  hat = (stab_list[1][max] - stab_list[0][max]);
	}   
	else	{
	  s = (stab_list[k][max] - stab_list[0][max]) / hat;
	  dist_list[k] = s;
	}
  }

  /* bubble sort the list by signed distance */
  for (k = 0; k < nplanes; k++)
	for (j = k+1; j < nplanes; j++)	{
	  if (dist_list[j] > dist_list[k])	{
		TSWAP (dist_list[j], dist_list[k], t);
		TSWAP (stab_list[j][0], stab_list[k][0], t);
		TSWAP (stab_list[j][1], stab_list[k][1], t);
		TSWAP (stab_list[j][2], stab_list[k][2], t);
		TSWAP (stab_list[j][3], stab_list[k][3], t);
	  }
	}

  /* connect the sorted points */
  for (k = 0; k < nplanes-1; k++)
	connect_p4 (stab_list[k], stab_list[k+1], k & 1 ? codd : ceven, k & 1 ? codd : ceven);
}

/* uses global draw_cylinders */
void
connect_p4 (
			float *A, float *B,
			long headc, long bodyc)
{
  float P[3];

  if (!(p_ideal (A, 4) || p_ideal (B, 4)))	{
	P[0] = A[0] * 0.2 + B[0] * 0.8;
	P[1] = A[1] * 0.2 + B[1] * 0.8;
	P[2] = A[2] * 0.2 + B[2] * 0.8;

	render_oriented_line ((POINT *)A, (POINT *)P, (POINT *)B, draw_cylinders, 0.5, bodyc, headc);
  }
}

int
draw_incident_lines()
{
  float A[4], B[4], C[4], D[4], L[4][6], T1[6], T2[6];
  int k, nsolutions;
  float N[3];

  for (k = 0; k < 4; k++)
	p_construct (segments[k][0], segments[k][1], segments[k][2], segments[k][3],
				 segments[k][4], segments[k][5], segments[k][6], segments[k][7], L[k]);

  nsolutions = p_incident4 (L[0], L[1], L[2], L[3], T1, T2);

  if (nsolutions == 1 || nsolutions == 2) {
	p_deconstruct (T1, A, B);
	connect_p4 (A, B, RGB_RED, RGB_CYAN);

	if (nsolutions == 2)	{
	  p_deconstruct (T2, C, D);
	  connect_p4 (C, D, RGB_RED, RGB_YELLOW);
	}

	if (debug && nsolutions == 2)  {	    /* print the lines to standard error */
	  fprintf (stderr, "line 0: [%8.4f %8.4f %8.4f] => [%8.4f %8.4f %8.4f]\n",
			   segments[0][0], segments[0][1], segments[0][2], 
			   segments[0][4], segments[0][5], segments[0][6]); 

	  fprintf (stderr, "line 1: [%8.4f %8.4f %8.4f] => [%8.4f %8.4f %8.4f]\n",
			   segments[1][0], segments[1][1], segments[1][2], 
			   segments[1][4], segments[1][5], segments[1][6]); 

	  fprintf (stderr, "line 2: [%8.4f %8.4f %8.4f] => [%8.4f %8.4f %8.4f]\n",
			   segments[2][0], segments[2][1], segments[2][2], 
			   segments[2][4], segments[2][5], segments[2][6]); 

	  fprintf (stderr, "line 3: [%8.4f %8.4f %8.4f] => [%8.4f %8.4f %8.4f]\n",
			   segments[3][0], segments[3][1], segments[3][2], 
			   segments[3][4], segments[3][5], segments[3][6]); 

	  fprintf (stderr, "soln 1: [%8.4f %8.4f %8.4f] => [%8.4f %8.4f %8.4f]\n",
			   A[0], A[1], A[2], B[0], B[1], B[2]);

	  fprintf (stderr, "soln 2: [%8.4f %8.4f %8.4f] => [%8.4f %8.4f %8.4f]\n",
			   C[0], C[1], C[2], D[0], D[1], D[2]);
	}
  }
  else if (nsolutions == 3)	{	/* Ti are a spanning basis */
	push_lighting (1);
	p_deconstruct (T1, A, B);
	p_deconstruct (T2, C, D);

	cross3f ((float)(B[0] - A[0]), (float)(B[1] - A[1]), (float)(B[2] - A[2]), 
			 (float)(D[0] - C[0]), (float)(D[1] - C[1]), (float)(D[2] - C[2]), 
			 N);

	CPACK (RGB_3);

	bgnpolygon ();
	n3f ((float *)N);
	v3f ((float *)A);
	v3f ((float *)B);
	v3f ((float *)C);
	endpolygon ();

	bgnpolygon ();
	n3f ((float *)N);
	v3f ((float *)A);
	v3f ((float *)B);
	v3f ((float *)D);
	endpolygon ();

	bgnpolygon ();
	n3f ((float *)N);
	v3f ((float *)B);
	v3f ((float *)C);
	v3f ((float *)D);
	endpolygon ();

	bgnpolygon ();
	n3f ((float *)N);
	v3f ((float *)A);
	v3f ((float *)C);
	v3f ((float *)D);
	endpolygon ();

	bgnpolygon ();
	n3f ((float *)N);
	v3f ((float *)A);
	v3f ((float *)C);
	v3f ((float *)D);
	endpolygon ();

	pop_lighting();
  }

  return (nsolutions);
}

handle_redraw()
{
}

int
DIRTY()
{
  int rval;

  return (dirty_data);
}

/* what to do if ASSERT() forces a core dump */
void
bailout(
		char *s,
		char *file,
		int line)
{
  fprintf (stderr, "ASSERTION FAILED, performing bailout action:\n");
  fprintf (stderr, "%s", s);
  snapshot_nd_data();
  _assert (s, file, line);
}

void
markfn (ARRAY *phalves)
{
  int k;
  PLUCKER *pl;

  for (k = 0; k < ALENGTH(phalves); k++)  {
	pl = (PLUCKER *)select_array (phalves, k);
	pl->special = CCOLOR (contours[pl->uid], pl->pid) ? 1 : 0;
  }
}

void
get_antipenumbral_events (
						  LIST *clist,	/* contour 0 == light source; others are holes */
						  int *ps,		/* effective problem size (computed from clist) */
						  LIST **events,	/* the returned event list */
						  int *nv, int *nf, int *nr, int *nc,		/* number of verts, faces, ridges, cliffs in computed 5d convex hull */
						  POINT *sP, POINT *sQ)   /* stabbing line, if successful */
{
  LIST *apevents, *lapevents, *auevents;

  assert (!LEMPTY(clist));
  assert (*events == NULL);

  /* linearize 0: do quadratic apen */
  /* linearize 1: do linear apen */
  /* linearize 2: do antiumbra */
  /* linearize 3: do antiumbra + linear apen */
  /* linearize 4: do quad + linear apen */
  /* linearize 5: do all three */
  *events = apevents = lapevents = auevents = NULL;

  /* quadric or both */
  if (linearize_ap == 0 || linearize_ap >= 4)
	compute_antipenumbra ((CONTOUR **) handle_list (clist, 0), LLENGTH(clist), 
						  ps, nv, nf, nr, nc, &apevents, markfn);

  if (linearize_ap == 1 || linearize_ap >= 3)	    {
	if (!testing_lap)
	  compute_linear_antipenumbra ((CONTOUR **) handle_list (clist, 0), LLENGTH(clist), ps, &lapevents, sP, sQ);
	else	{
	  // #define TEST_INCREMENTAL
#ifndef TEST_INCREMENTAL
	  LIST *oldlapevents;

	  // test -- first compute ap of first n-1 contours
	  oldlapevents = NULL;
	  if (LLENGTH(clist) > 1) 
		compute_linear_antipenumbra ((CONTOUR **) handle_list (clist, 0), 
									 LLENGTH(clist)-1, ps, &oldlapevents, sP, sQ);

	  // now tack on final one but re-use old list
	  compute2_linear_antipenumbra ((CONTOUR **) handle_list (clist, 0),
									LLENGTH(clist), ps, oldlapevents, &lapevents, sP, sQ);

	  apen_discard_event_list (&oldlapevents);
#else	
	  // test incremental ap alg
	  CONTOUR *c, *cA, *cZ;
	  LIST *nclist;
	  int k;

	  extern int ilap_update (CONTOUR *, LIST **, LIST **, CONTOUR **, CONTOUR **);
	    
	  nclist = NULL;
	  lapevents = NULL;
	  for (k = 0; k < LLENGTH(clist); k++)    {
		c = (CONTOUR *)select_list (clist, k);
		if (!ilap_update (c, &nclist, &lapevents, &cA, &cZ))
		  break;

		// draw cA, cZ; then delete them?
		discard_contour (&cA);
		discard_contour (&cZ);
	  }

	  discard_list (&nclist);
#endif
	}
  }

  if (linearize_ap == 2 || linearize_ap == 3 || linearize_ap >= 5)
	compute_antiumbra ((CONTOUR **) handle_list (clist, 0), LLENGTH(clist), ps, &auevents, sP, sQ);

  if (apevents != NULL)
	*events = apevents;

  if (lapevents != NULL)  {
	if (*events == NULL)
	  *events = lapevents;
	else
	  append_list (events, &lapevents);
  }

  if (auevents != NULL)  {
	if (*events == NULL)
	  *events = auevents;
	else
	  append_list (events, &auevents);
  }
}

/* extends contour list by one element, then rolls it back */
void
draw_antipenumbral_events (
						   LIST *clist,		/* contour 0 == light source; others are holes */
						   LIST *events,		/* events collection */
						   int eventmask,		/* event selection */
						   PLANE *farp,		/* far plane if non-NULL */
						   int drawaux)		/* draw projection plane */
{
  PLANE *near, *far;
  int k, highres, n_eeee;
  CONTOUR *c0, *dc;

  c0 = (CONTOUR *) head_list (clist);

  if (LLENGTH(clist) > 1) {
	/* extend contour list by one with a translated doppleganger */
	dc = dopple_contour (select_list (clist, LLENGTH(clist) - 1), -cextrusion_distance);

	insert_list (clist, (void *)dc);

	near = CPLANE(((CONTOUR *)select_list (clist, LLENGTH(clist)-2)));
	far  = (farp != NULL) ? farp : CPLANE(((CONTOUR *)select_list (clist, LLENGTH(clist)-1)));

	{
	  // highres = script_playback_file || !(leftdown || middledown || rightdown);
	  highres = TRUE;
	}
	
	if (events != NULL)
	  draw_penumbral_events (events, eventmask, near, far,
							 (CONTOUR **)handle_list (clist, 0), LLENGTH(clist),
							 w_density, highres,
							 /* RGB_GREEN, RGB_PURPLE, RGB_BLUE, */  RGB_GREEN, RGB_GREEN, RGB_BLUE,

							 fixed_color_table[hackcolor].cid /* RGB_RED */, /* RGB_CHOCOLATE, RGB_MAROON, */ RGB_RED, RGB_RED,

							 RGB_LIGHTBLUE, RGB_SEAGREEN);

	if (events != NULL && !(eventmask & EEEE_EVENT_TYPE) && pref_seg >= 0)	{
	  EVENT *event;
	  LIST *telist;

	  /* find the kth EEEE event and display it */
	  n_eeee = 0;
	  for (k = 0; k < LLENGTH(events); k++)   {
		event = (EVENT *)select_list (events, k);
		if (event->etype == EEEE_EVENT_TYPE)	{
		  if (n_eeee++ == pref_seg)	{
			telist = create_list (sizeof (EVENT *));
			insert_list (telist, (void *)event);
			draw_penumbral_events (telist, EEEE_EVENT_TYPE, near, far,
								   (CONTOUR **)handle_list (clist, 0), LLENGTH(clist),
								   w_density, highres,
								   RGB_GREEN, RGB_GREEN, RGB_BLUE,
								   fixed_color_table[hackcolor].cid /* RGB_RED */, 
								   /* RGB_CHOCOLATE, RGB_MAROON, */ RGB_RED, RGB_RED,
								   RGB_LIGHTBLUE, RGB_SEAGREEN);
			discard_list (&telist);
			break;
		  }
		}
	  }
	}

	if (suppressing_contours)
	  return;
	    
	/* draw holes */
	for (k = 1; k < LLENGTH(clist)-1; k++)
	  draw_contour (contours[k], FALSE, FALSE, highres,
					(k != pref_contour) ? RGB_CYAN : RGB_7, (k != pref_contour) ? RGB_DARKBLUE : RGB_ORANGE);

	/* draw aux hole */
	if (drawaux)
	  draw_contour ((CONTOUR *) tail_list (clist), FALSE, FALSE, highres, RGB_ABYSS, RGB_ABYSS);

	/* discard the doppled contour */
#if 0
	discard_contour ((CONTOUR **) handle_list (clist, LLENGTH(clist)-1));
#else
	discard_contour ((CONTOUR **) &dc);
#endif

	/* remove its place on list */
	truncate_list (clist, LLENGTH(clist)-1);
  }
  else    {
	PLANE negplane;

	/* invoke a h/w clipping plane along the zeroth contour */
	/* draw a white hemisphere centered on the contour */
	push_lighting(1);
	MPACK (RGB_LIGHTGREY);
	negplane = *CPLANE(c0);
	negate_plane (&negplane);
	clipplane (0, CP_DEFINE, (float *)&negplane);
	clipplane (0, CP_ON, NULL);
	draw_sphere (CCENTROID(c0), contour_diameter(c0));
	clipplane (0, CP_OFF, NULL);
	pop_lighting();
  }

  /* fill light source */
  if (fill_light_source && !show_exploded)  {
	CPACK (printscreen ? RGB_GREY : RGB_WHITE);
	fill_contour (c0);
  }

  /* draw light source */
  draw_contour (c0, FALSE, FALSE, highres, RGB_WHITE, RGB_ORANGE);
}

/* takes a contour of length n, returns n quadrilaterals in list, oppositely directed */
LIST *
explode_contour (
				 CONTOUR *c)
{
  CONTOUR *d, *e;
  LIST *clist;
  int k;
  POINT *cv, *dv, *ev;

  d = copy_contour (c);
  scale_contour (d, 5.0 + 3.0 * cextrusion_distance);

  cv = CVERTICES(c);
  dv = CVERTICES(d);
  clist = create_list (sizeof (CONTOUR *));
  for (k = 0; k < CLENGTH(c); k++)	{
	e = create_dummy_colored_contour (4);
	insert_list (clist, (void *) e);
	ev = CVERTICES(e);

	*(ev + 0) = *(cv + (k+1) % CLENGTH(c));
	*(ev + 1) = *(cv + k); 
	*(ev + 2) = *(dv + k);
	*(ev + 3) = *(dv + (k+1) % CLENGTH(d));

	find_contour_attributes (e);
  }

  discard_contour (&d);
  return (clist);
}

/* returns 0 on overflow */
int
bump_array (
			ARRAY *indexa)
{
  int k, carry = 0;
  int *indexp, *imax;


  for (k = 1; k < ALENGTH(indexa) / 2; k++)   {
	indexp = (int *)select_array (indexa, 2 * k);
	imax = (int *)select_array (indexa, 2 * k + 1);
	if (++(*indexp) == *imax)
	  *indexp = 0;
	else
	  break;
  }

  return (k < ALENGTH(indexa) / 2);
}

void
get_exploded_contour_events (
							 LIST *clist, 
							 LIST **events)
{
  LIST *ecl_list;	    /* list of lists */
  LIST *eclist;	    /* list of lists */
  ARRAY *indexa;	    /* array of indices */
  LIST *pathlist;	    /* list of CONTOURs */
  int k, ps, nv, nf, nr, nc, *indexp;
  static LIST *pevents = NULL;
  int zero = 0, imax;
  EVENT *event;
  VE_EVENT *ve_event;
  EEE_EVENT *eee_event;
  EEEE_EVENT *eeee_event;

  assert (!LEMPTY(clist));
  assert (*events == NULL);

  ecl_list = create_list (sizeof (LIST *));
  indexa = create_array (sizeof (int));

  insert_list (ecl_list, NULL);
  insert_array (indexa, (void *)&zero);
  insert_array (indexa, (void *)&zero);
  for (k = 1; k < LLENGTH(clist); k++)    {
	eclist = explode_contour ((CONTOUR *)select_list (clist, k));
	insert_list (ecl_list, (void *)eclist);
	imax = LLENGTH(eclist);
	insert_array (indexa, (void *)&zero);   /* initialize index */
	insert_array (indexa, (void *)&imax);   /* initialize max index */
  }

  /* now loop over all permutations of holes */
  pathlist = create_list (sizeof (CONTOUR *));
  insert_list (pathlist, (void *) head_list (clist));	/* light source always contour 0 */

  *events = create_list (sizeof (EVENT *));

  do	{
	truncate_list (pathlist, 1);
	for (k = 1; k < LLENGTH(clist); k++)    {
	  eclist = select_list (ecl_list, k);
	  indexp = (int *) select_array (indexa, 2 * k);
	  insert_list (pathlist, (void *) select_list (eclist, *indexp));
	}

	/* this call allocates and returns the events array */
	compute_antipenumbra ((CONTOUR **) handle_list (pathlist, 0), LLENGTH(pathlist),
						  &ps, &nv, &nf, &nr, &nc, &pevents, NULL);

	/* grab the generated events: if involve hole edge, must be edge 0 */
	for (k = 0; k < LLENGTH(pevents); k++)	{
	  event = (EVENT *) select_list (pevents, k);

#define FILTER
#ifdef FILTER
	  switch (event->etype)	{
	  case PVE_EVENT_TYPE :
	  case UVE_EVENT_TYPE :
	  case IVE_EVENT_TYPE :
		ve_event = (VE_EVENT *)event;
		/* A->pid2, B->pid are shared vertex */
		assert (ve_event->A->pid2 == ve_event->B->pid);

		/* if [light edge x hole vertex], penumbral and internal events must be swapped */
		if (ve_event->A->uid > 0)   {
		  if (event->etype == PVE_EVENT_TYPE)
			event->etype = IVE_EVENT_TYPE;
		}

		if (((ve_event->A->uid == 0) || (ve_event->A->pid2 + ve_event->B->pid <= 2)) &&
		    ((ve_event->C->uid == 0) || (ve_event->C->pid + ve_event->C->pid2 == 1)))
		  insert_list (*events, (void *)event);
		break;

	  case PEEE_EVENT_TYPE :
	  case IEEE_EVENT_TYPE :
		eee_event = (EEE_EVENT *)event;
		if (((eee_event->A->uid == 0) || (eee_event->A->pid + eee_event->A->pid2 == 1)) &&
		    ((eee_event->B->uid == 0) || (eee_event->B->pid + eee_event->B->pid2 == 1)) &&
		    ((eee_event->C->uid == 0) || (eee_event->C->pid + eee_event->C->pid2 == 1)))
		  insert_list (*events, (void *)event);
		break;

	  case EEEE_EVENT_TYPE :
		eeee_event = (EEEE_EVENT *)event;
		if (((eeee_event->A->uid == 0) || (eeee_event->A->pid + eeee_event->A->pid2 == 1)) &&
		    ((eeee_event->B->uid == 0) || (eeee_event->B->pid + eeee_event->B->pid2 == 1)) &&
		    ((eeee_event->C->uid == 0) || (eeee_event->C->pid + eeee_event->C->pid2 == 1)) &&
		    ((eeee_event->D->uid == 0) || (eeee_event->D->pid + eeee_event->D->pid2 == 1)))
		  insert_list (*events, (void *)event);
		break;

	  default:
		break;
	  }
#else
	  insert_list (*events, (void *)event);
#endif
	}

	discard_list (&pevents);
	/* NOTE -- SOMEONE MUST FREE FILTERED EVENTS */
	/*   	apen_discard_event_list (&pevents);	*/
  } while (bump_array (indexa));

  discard_array (&indexa);
  apply_list_byref (ecl_list, discard_list);
  discard_list (&ecl_list);
}

void
draw_exploded_contours (
						LIST *clist)
{
  LIST *eclist;
  int k, j;

  for (k = 1; k < LLENGTH(clist); k++)    {
	// draw_filled_contour ((CONTOUR *)select_list (clist, k), RGB_CHOCOLATE);
	eclist = explode_contour ((CONTOUR *)select_list (clist, k));
	for (j = 0; j < LLENGTH(eclist); j++)
	  draw_contour ((CONTOUR *) select_list (eclist, j), FALSE, FALSE, FALSE, RGB_YELLOW, RGB_CHOCOLATE);
	apply_list_byref (eclist, discard_contour);
	discard_list (&eclist);
  }
}

void
fill_exploded_contours (
						LIST *clist)
{
  LIST *eclist;
  int k, j;
  CONTOUR *ck, *cj;

  for (k = 1; k < LLENGTH(clist); k++)    {
	ck = (CONTOUR *) select_list (clist, k);
	// draw_contour (ck, FALSE, FALSE, FALSE, RGB_YELLOW, RGB_DARKGREEN);

	eclist = explode_contour (ck);
	for (j = 0; j < LLENGTH(eclist); j++)	{
	  cj = (CONTOUR *)select_list (eclist, j);
	  draw_filled_contour (cj, RGB_BROWN4);
	}

	apply_list_byref (eclist, discard_contour);
	discard_list (&eclist);
  }
}

void
my_swapbuffers(void)
{
  swapbuffers();

  {
    int iomode;
    extern char *script_playback_file, *script_record_file;
    char title[128];

    iomode = tape_getmode();
    if (iomode == TAPE_PLAYBACK)    {
	  sprintf (title, "playing script %s, frame %5d\n", script_playback_file, frame_num);
	  wintitle (title);
	}
    else if (iomode == TAPE_RECORD)  {
	  sprintf (title, "recording script %s, frame %5d\n", script_record_file, frame_num);
	  wintitle (title);
	}
  }

  if (printscreen)	{
	winsave ();
	printscreen = 0;
  }

  if (writing_frames)	{
	char *getenv(const char *);
	static char sfname[32];
	static char *tmpdir;

	if (frame_num == 0)
	  tmpdir = getenv ("TMPDIR");

	if ((skip_frame <= 0 || !(frame_num % skip_frame)) && first_frame <= frame_num && frame_num <= last_frame)    {
	  if (tmpdir != NULL)
		sprintf (sfname, "%s/frame_%04d.rgb", tmpdir, frame_num);
	  else
		sprintf (sfname, "frame_%04d.rgb", frame_num);
#define FWRITE_ENABLED
#ifdef FWRITE_ENABLED
	  savewindow (sfname);
#else
	  fprintf (stderr, "writing frame '%s'\n", sfname);
#endif
	}
  }

  frame_num++;
}

void
draw_p3bbox (Point3 *pmin, Point3 *pmax)
{
  EXTENT extent;

  extent[0][0] = pmin->x; 
  extent[0][1] = pmin->y; 
  extent[0][2] = pmin->z; 
  extent[1][0] = pmax->x; 
  extent[1][1] = pmax->y; 
  extent[1][2] = pmax->z; 
  outline_extent_3D (extent);
}

// given a point p, compute solid angle of global contour
static CONTOUR *global_ls;
static CONTOUR **global_hlist;
static int global_nholes;

// given a sample point, compute solid angle of light
// as seen from point.  use hole list to clip light
// source to region visible through holes
float 
sol_ang (POINT *p)
{
  CONTOUR *inclip, *outclip, *hole;
  int hid, eid;
  float rval;
  POINT *hp, *hq;
  PLANE Peqn;

  // first clip the light source to each hole (projected through p)
  inclip = copy_contour (global_ls);
  for (hid = 0; hid < global_nholes; hid++)	{
	hole = global_hlist[hid];
	hq = CVERTEX(hole, CLENGTH(hole)-1);
	for (eid = 0; eid < CLENGTH(hole); eid++)   {
	  hp = hq;
	  hq = CVERTEX(hole, eid);
	  // generate plane spanned by p and contour edge, containing centroid
	  plane_through_three_points_below_point (&Peqn, p, hp, hq, CCENTROID(hole));
	    
	  // clip current contour to this plane
	  clip_contour (inclip, &Peqn, &outclip);
	  discard_contour (&inclip);
	  if (outclip == NULL)
		return (0.0);
	  else
		inclip = outclip;
	}
  }

  rval = convex_ngon_solid_angle (p, CVERTICES(inclip), CLENGTH(inclip)) / (4.0 * PI);
  discard_contour (&inclip);

  // square it to accentuate slope discontinuities?

  return (rval);
}

void
graph_solidangle (
				  CONTOUR *light,
				  CONTOUR **holes, int nholes,
				  CONTOUR *domain,
				  int density,
				  int flush,
				  long colorid)
{
  global_ls = light;
  global_hlist = holes;
  global_nholes = nholes;

  compute_draw_function (CCENTROID(domain), CNORMAL(domain), C_SHAT(domain), C_THAT(domain),
						 -shadow_length, -shadow_length, shadow_length, shadow_length, 1.0, density, flush, sol_ang, NULL);
}

// assumes disk is coplanar with C0
// basic idea: for each pair of contours Ci, Cj, i <> j, i,j > 0,
// generate the swaths due to disk and all contour edges
// clip these swaths to contours Ck, k <> i, k <> j, and draw them
void
draw_disk_swaths (
				  POINT *dcenter,
				  PLANE *dplane,
				  float dradius,
				  CONTOUR **contours,
				  int ncontours,
				  int density)
{
  int i, j, k, ei, ej, ek, d;
  float angle;
  CONTOUR *ci, *cj;
  static PLANE Farp, Ppqr, *far_pl = &Farp, *near_pl;
  static POINT Pclip, *pclip = &Pclip;
  static POINT Pnear, *pnear = &Pnear, Pfar, *pfar = &Pfar;
  static POINT Pnear0, *pnear0 = &Pnear0, Pfar0, *pfar0 = &Pfar0;
  static VECTOR Shat, That, *shat = &Shat, *that = &That;
  static POINT Lead, *lead = &Lead;
  static POINT ELead, *elead = &ELead;
  int valid;
    
  construct_orthobasis ((VECTOR *)dplane, shat, that);
  vscale (shat, dradius);
  vscale (that, dradius);

  near_pl = CPLANE (contours[ncontours-1]);
  plane_from_plane_and_distance (far_pl, CPLANE (contours[ncontours-1]), -cextrusion_distance);

  for (i = 1; i < ncontours; i++) {
	ci = contours[i];
	for (j = i; j < ncontours; j++)   {
	  cj = contours[j];
	  for (ei = 0; ei < CLENGTH(ci); ei++)
		for (ej = 0; ej < CLENGTH(cj); ej++)	{
		  if (i == j)	{
			if (ei != ej)
			  continue;
			if (!show_halfspaces)
			  continue;
			MPACK (RGB_BLUE);
		  }
		  else    {
			if (!show_antipenumbra)
			  continue;
			MPACK (RGB_RED);
		  }

		  // sweep out curve due to circle partition
		  valid = 0;
		  for (d = 0; d <= density; d++)   {
			*pnear0 = *pnear;
			*pfar0 = *pfar;

			// generate point lead on circle
			angle = (d * TWO_PI) / density;
			(void)pplus_su_tv (lead, dcenter, fcos (angle), shat, fsin(angle), that);

			if (i == j) {	    // circle / VERTEX event
			  // elead is just vertex ci of contour i (or j)
			  *elead = *CVERTEX(ci, ei);
			}
			else	{	    // circle / EDGE / EDGE event
			  // compute extrapolated point elead
			  // first find plane through lead and edge ei
			  plane_through_three_points (&Ppqr, lead, CVERTEX(ci, ei), CVERTEX(ci, (ei+1)%CLENGTH(ci)));
    
			  // intersect this plane with the edge ej
			  intersect_line_plane (elead, CVERTEX(cj, ej), CVERTEX(cj, (ej+1)%CLENGTH(cj)), &Ppqr);
			
			  if (!point_between (elead, CVERTEX(cj, ej), CVERTEX(cj, (ej+1) % CLENGTH(cj))))	{
				valid = 0;
				continue;
			  }

			  // and with plane of contour ci, for clip check
			  intersect_line_plane (pclip, lead, elead, CPLANE(ci));

			  if (!point_between (pclip, CVERTEX(ci, ei), CVERTEX(ci, (ei+1) % CLENGTH(ci))))	{
				valid = 0;
				continue;
			  }
			}

			// check line against all contours k > 0, k <> i, j
			for (k = 1; k < ncontours; k++)	{
			  if (k == i || k == j)
				continue;
			  if (!line_intersect_contour (pclip, lead, elead, contours[k]))
				break;
			}

			if (k < ncontours)  {
			  valid = 0;
			  continue;
			}

			// finally, intersect with near, far planes and draw		
			intersect_line_plane (pnear, lead, elead, near_pl);
			intersect_line_plane (pfar, lead, elead, far_pl);

			if (valid)  {	    // pnear0, pfar0 valid
			  // uncapped_cylinder (pnear, pfar, dradius * 0.02, 5);
			  face3p (pnear0, pfar0, pfar);
			  face3p (pnear0, pfar, pnear);
			}

			valid = 1;
		  }
		}
	}
  }
}

/* master draw routine */
void
REFRESH_ns( void )
{
  int j, k, event_types;
  static int attempts = 1;
  static int nsolns, ps, nv, nf, nr, nc;
  static PLUCKER P, Q;
  float A[4], B[4];
  char str[80], str2[80];
  static LIST *clist = NULL, *rclist = NULL;
  static LIST *events = NULL, *revents = NULL;
  POINT sP, sQ, sP2, sQ2, pF;
  static PLANE Farp, *farp = NULL;
  VECTOR vPQ;
  static int old_ts = -1;
  static int vis = 0;

  if (spinning)   {
	xyrotate (2.0, 2.0);
  }

  if (!do_blockers)
	set_viewing(printscreen ? -1 : clearcolor);	    /* function of sites bbox */
  else 	
	set_viewing(clearcolor);	    /* function of sites bbox */
#ifndef PSGL
  zbuffer (TRUE);
#else
  zbuffer (!psgl_enabled);
#endif

  sphmode (SPH_DEPTH, sphere_depth);
  if (old_ts != show_twosided)	{
	twosided_lighting (show_twosided);
	old_ts = show_twosided;
  }

  iomode = tape_getmode();
  if (control && iomode != TAPE_PLAYBACK)    {
	draw_unit_cube_wires (0);
	draw_unit_cube_wires (1);
	draw_unit_cube_wires (2);
  }

  /* bind rgb colors to current material */
  /* lmcolor (LMC_AD); */
  linewidth ((int)line_width);

  if (!script_playback_file && pref_face >= 0)    {
	zfunction (ZF_LESS);
	draw_unit_cube (pref_face);
	zfunction (ZF_LEQUAL);
  }

  /* compute event mask */
  event_types = 0;
  if (show_stabbers && pref_seg < 0)
	event_types |= EEEE_EVENT_TYPE;
  if (show_antipenumbra)
	event_types |= PEEE_EVENT_TYPE;
  if (show_ieee)
	event_types |= IEEE_EVENT_TYPE;
  if (show_halfspaces)
	event_types |= PVE_EVENT_TYPE;
  if (show_ive)
	event_types |= IVE_EVENT_TYPE;
  if (show_antiumbra)
	event_types |= UVE_EVENT_TYPE;

  if (experimental)	{

	if (linearize_ap == 1)	    /* linear only */
	  dirty_data = 1;

	/* assemble contours into list */
	clist = create_list (sizeof (CONTOUR *));
	rclist = create_list (sizeof (CONTOUR *));
	for (k = 0; k < ncontours; k++)	{
	  insert_list (clist, (void *)contours[k]);
	  if (twoway)
		insert_list (rclist, (void *)reverse_contour (contours[ncontours - k - 1]));
	}

	if (dirty_data) {
	  apen_discard_event_list (&events);
	  get_antipenumbral_events (clist, &ps, &events, &nv, &nf, &nr, &nc, &sP, &sQ);

	  if (twoway)	{
		apen_discard_event_list (&revents);
		get_antipenumbral_events (rclist, &ps, &revents, &nv, &nf, &nr, &nc, &sP2, &sQ2);
	  }
    
	  if (linearize_ap && !LNULL_OR_EMPTY (events))  {
#if 0
		vdiffp (&vPQ, &sQ, &sP);
		pplustv (&pF, &sQ, 0.5, &vPQ);
		plane_from_normal_and_point (&Farp, &vPQ, &pF);
		farp = &Farp;
#else
		farp = NULL;
#endif
	  }
	  else
		farp = NULL;

	  dirty_data = FALSE;
	}

	if (show_exploded)  {
	  push_lighting (0);
	  draw_filled_contour (head_list (clist), printscreen ? RGB_GREY : RGB_WHITE);
	  fill_exploded_contours (clist);
	  pop_lighting ();
	}

	draw_antipenumbral_events (clist, events, event_types, farp, 
							   control && iomode != TAPE_PLAYBACK);
	if (twoway)
	  draw_antipenumbral_events (rclist, revents, event_types, farp, 
								 control && iomode != TAPE_PLAYBACK);

	if (pref_hole >= 0 && pref_hole < ncontours)  {
	  CONTOUR *c;
	  POINT *a, *b, Mid;
	    
	  c = (CONTOUR *)select_list (clist, pref_hole);
	  if (pref_seg >= 0 && pref_seg < CLENGTH(c))	{
		push_lighting (1);
		a = CVERTEX(c, pref_seg);
		b = CVERTEX(c, (pref_seg + 1) % CLENGTH(c));
		plerp (&Mid, a, b, 0.8);
		render_oriented_line (a, &Mid, b,
							  TRUE, 1.25, RGB_MAGENTA, RGB_MAGENTA);
		pop_lighting ();
	  }
	}

	discard_list (&clist);
	if (twoway)
	  apply_list_byref (rclist, discard_contour);
	discard_list (&rclist);

	if (control && !shifted && tape_getmode() != TAPE_PLAYBACK) {
	  /* draw the origin */
	  linewidth (5);
	  CPACK (RGB_PINK);
	  move (0.0, 0.0, 0.0);
	  draw (0.0, 0.0, 0.0);
	  linewidth (0);
	}

	goto report;
  }
  else if (do_extents && ncontours == 2)  {
	PLANE sep[16], com[16], shaft[32];

	push_lighting (1);
	draw_contour (contours[0], 0, 0, 1, RGB_GREEN, RGB_YELLOW);
	draw_contour (contours[1], 0, 0, 1, RGB_BLUE, RGB_YELLOW);
	// generate_extent_separating_planes (PEXTENT (CBBOX(contours[0])), PEXTENT (CBBOX(contours[1])), sep);
	// generate_extent_common_planes (PEXTENT (CBBOX(contours[0])), PEXTENT (CBBOX(contours[1])), com);
	generate_extent_shaft_planes (PEXTENT (CBBOX(contours[0])), PEXTENT (CBBOX(contours[1])), shaft);
	pop_lighting();
  }
  else if (do_blockers)   {
	PQUAD *S, *R, *B;
	TUBE *tube;
	int cid;

	// contours 0 and 1 are source; 2 .. n-1 are blockers
	for (cid = 0; cid < ncontours; cid++)	    
	  assert (CLENGTH(contours[cid]) == 4);

	if (subdivideS >= 0)	{
	  subd_contour (contours[0], subdivideS);
	  subdivideS = -1;
	}

	if (subdivideR >= 0)	{
	  subd_contour (contours[ncontours-1], subdivideR);
	  subdivideR = -1;
	}

	// generate PQUADs for each contour
	S = pl_alloc_pquad (
						(Point3 *)CVERTEX(contours[0], 0), 
						(Point3 *)CVERTEX(contours[0], 1), 
						(Point3 *)CVERTEX(contours[0], 2), 
						(Point3 *)CVERTEX(contours[0], 3),
						(HPlane3 *)CPLANE(contours[0]));

	R = pl_alloc_pquad (
						(Point3 *)CVERTEX(contours[ncontours-1], 0), 
						(Point3 *)CVERTEX(contours[ncontours-1], 1), 
						(Point3 *)CVERTEX(contours[ncontours-1], 2), 
						(Point3 *)CVERTEX(contours[ncontours-1], 3),
						(HPlane3 *)CPLANE(contours[ncontours-1]));

	tube = pl_alloc_tube (S, R, NULL, ncontours-2);

	tube->nb = 0;
	for (cid = 1; cid < ncontours-1; cid++)	{
	  B = pl_alloc_pquad (
						  (Point3 *)CVERTEX(contours[cid], 0),
						  (Point3 *)CVERTEX(contours[cid], 1),
						  (Point3 *)CVERTEX(contours[cid], 2),
						  (Point3 *)CVERTEX(contours[cid], 3),
						  (HPlane3 *)CPLANE(contours[cid]));
	  /*
	    if (!pl_tubeRS_incident_blocker (tube, B, 1))   {
		pl_free_pquad (&B);
		B = NULL;
		continue;
		}
		*/
	  if (tube->B)
		tube->B[tube->nb++] = B;
	}

	// complain if tube is bad
	pl_verify_tube (tube, 1);

	push_lighting (1);
	draw_contour (contours[0], 0, 0, 1, RGB_GREEN, RGB_YELLOW);
	if (fill_light_source)	{
	  MPACK (RGB_GREEN);
	  box_slabs (CVERTEX(contours[0], 0), CVERTEX(contours[0], 1), 
				 CVERTEX(contours[0], 2), CVERTEX(contours[0], 3),
				 CNORMAL(contours[0]),
				 apen_cyl_radius * 0.6, 16);
	}
	draw_contour (contours[ncontours-1], 0, 0, 1, RGB_GREEN, RGB_YELLOW);
	if (fill_light_source)	{
	  MPACK (RGB_GREEN);
	  box_slabs (CVERTEX(contours[ncontours-1], 0), CVERTEX(contours[ncontours-1], 1), 
				 CVERTEX(contours[ncontours-1], 2), CVERTEX(contours[ncontours-1], 3),
				 CNORMAL(contours[ncontours-1]),
				 apen_cyl_radius * 0.6, 16);
	}
	pop_lighting();

	// draw bbox
	if (debug)  {
	  CPACK (RGB_CYAN);
	  draw_p3bbox (&tube->pmin, &tube->pmax);
	}

	if (do_blockers == 1) {
	  for (k = 0; k < tube->nb; k++)  {
		B = tube->B[k];

		// classify S,R as VISIBLE, INVISIBLE, or PARTIAL w.r.t. B !
		if (tube->vis == INVISIBLE)
		  vis = INVISIBLE;
		else    {
		  // draw bbox
		  if (debug && B)  {
			CPACK (RGB_MAGENTA);
			draw_p3bbox (&B->pmin, &B->pmax);
		  }
		}

		// how to handle multiple contour case... ?
		vis = pl_classify_visibility (S, R, tube->cross, tube->ncross, B);

		// draw the blocker
		push_lighting(1);

		// draw vis in RED (INVISIBLE) ; WHITE (VISIBLE) ; BLUE (PARTIAL)
		if (vis == INVISIBLE)
		  MPACK (RGB_BLUE);
		else if (vis == VISIBLE)
		  MPACK (RGB_CYAN);
		else if (vis == PARTIAL_HARD)
		  MPACK (RGB_RED);
		else if (vis == VISIBLE_HARD)
		  MPACK (RGB_GREEN);
		else	{
		  assert (vis == PARTIAL);
		  MPACK (RGB_MAGENTA);
		}

		//		rounded_pbox(....);
		box_slabs (CVERTEX(contours[k+1], 0), CVERTEX(contours[k+1], 1), 
				   CVERTEX(contours[k+1], 2), CVERTEX(contours[k+1], 3),
				   CNORMAL(contours[k+1]),
				   apen_cyl_radius * 0.6, 16);

		draw_contour (contours[k+1], 0, 0, 1, RGB_YELLOW, RGB_YELLOW);
	    
		pop_lighting();
	  }

	  push_lighting(1);
	  // draw the cross-talk
	  for (k = 0; k < tube->ncross; k++)	{
		// classify the wire wrt the blocker(s)
		if (pl_classify_segment (tube->cross[k].s, tube->cross[k].r, tube))
		  MPACK (RGB_RED);	    // hit
		else
		  MPACK (RGB_WHITE);	    // miss

		uncapped_cylinder ((POINT *)tube->cross[k].s, (POINT *)tube->cross[k].r, apen_cyl_radius, 16);
	  }
	  pop_lighting ();
	}
	else	{
	  static int db_0 = -1;
	  static int nhits, hits[256];
	  static POINT randp[256], randq[256];

	  if (do_blockers != db_0)	{
		db_0 = do_blockers;
		// generate random rays
		for (k = 0; k < do_blockers; k++)   {
		  contour_random_point (contours[0], &randp[k]);
		  contour_random_point (contours[1], &randq[k]);
		}
	  }

	  // sample the blockers versus the tube
	  for (k = nhits = 0; k < do_blockers; k++)	{
		if (pl_classify_segment ((Point3 *)&randp[k], (Point3 *)&randq[k], tube)) {
		  hits[k] = 1;
		  nhits++;
		}
		else    {
		  hits[k] = 0;
		}
	  }

	  push_lighting (0);
	  for (k = 0; k < do_blockers; k++)   {
		if (hits[k])
		  CPACK (RGB_RED);
		else
		  CPACK (RGB_GREEN);
		wire2p (&randp[k], &randq[k]);
	  }
	  pop_lighting ();

	  push_lighting(1);
	  // draw the blocker, shaded appropriately
	  MPACKCPD (lerp_color (0x0, 0xFFFFFFFF, (do_blockers - nhits) / (float)do_blockers));
	  fill_contour (contours[2]);
	  pop_lighting ();
	}

	// clean up
	pl_free_tube (&tube);
	pl_free_pquad (&B);
	pl_free_pquad (&R);
	pl_free_pquad (&S);
  }
  else if (do_ff) {
	push_lighting(1);
	// draw the contours
	for (k = 0; k < ncontours; k++)
	  draw_contour (contours[k], 0, 0, 1, RGB_GREEN, RGB_YELLOW);

#define CT(cid) contours[cid]
#define CV(cont,vid) CVERTEX(cont, vid)
#define cs contour_separating_tangent_plane
	{
	  CONTOUR *A, *B;
	  POINT VA, *va = &VA;
	  POINT VB, *vb = &VB;
	  PLANE sep01, sep12, sep20;

	  // should be two triangles
	  // draw the three separating planes from A to B, and B to A
	  A = CT(0);
	  B = CT(1);

	  MPACK (RGB_ORANGE);

	  cs (A, CV(A,0), CV(A,1), B, 0, &sep01, vb);
	  face3p (CV(A,0), CV(A,1), vb);

	  cs (A, CV(A,1), CV(A,2), B, 0, &sep12, vb);
	  face3p (CV(A,1), CV(A,2), vb);

	  cs (A, CV(A,2), CV(A,0), B, 0, &sep20, vb);
	  face3p (CV(A,2), CV(A,0), vb);

	  MPACK (RGB_BLUE);

	  cs (B, CV(B,0), CV(B,1), A, 1, &sep01, va);
	  face3p (CV(B,1), CV(B,0), va);

	  cs (B, CV(B,1), CV(B,2), A, 1, &sep12, va);
	  face3p (CV(B,2), CV(B,1), va);

	  cs (B, CV(B,2), CV(B,0), A, 1, &sep20, va);
	  face3p (CV(B,0), CV(B,2), va);

	  pop_lighting();

	  // compute the form factor
	  {   
		int k;
		double Averts[3][3];
		double Bverts[3][3];
	    
		for (k = 0; k < 3; k++)	{
		  Averts[k][0] = CV(A,k)->px;
		  Averts[k][1] = CV(A,k)->py;
		  Averts[k][2] = CV(A,k)->pz;

		  Bverts[k][0] = CV(B,k)->px;
		  Bverts[k][1] = CV(B,k)->py;
		  Bverts[k][2] = CV(B,k)->pz;
		}
		form_factor = FormFactor (Averts, 3, Bverts, 3);
	  }

	}

  }
  else if (do_disk_light) {
	draw_line_heads = 0;

	push_lighting(1);
	MPACK (RGB_CYAN);
	// treat first contour as its circumcircle
	cyl_circle3d (CCENTROID(contours[0]), CNORMAL(contours[0]), 
				  fsqrt (CAREA(contours[0])), 0.025 * fsqrt (CAREA(contours[0]))); 
	
	// draw the contours
	for (k = 0; k < ncontours; k++)
	  draw_contour (contours[k], 0, 0, 1, RGB_GREEN, RGB_YELLOW);

	// draw the swaths generated by light source, contour edges
	draw_disk_swaths (CCENTROID(contours[0]), CPLANE(contours[0]), fsqrt (CAREA(contours[0])),
					  contours, ncontours, r_density);

	pop_lighting();
  }
  else if (do_solidangle) {
	CONTOUR *clipc;

	// draw the contours
	for (k = 0; k < ncontours; k++)
	  draw_contour (contours[k], 0, 0, 1, RGB_GREEN, RGB_YELLOW);

#if 0	// CLIP test
	// draw the contours, clipped to one another
	if (clip_contour (contours[0], CPLANE(contours[1]), &clipc) == 1)   {
	  draw_contour (clipc, 0, 0, 1, RGB_RED, RGB_YELLOW);
	  discard_contour (&clipc);
	}

	if (clip_contour (contours[1], CPLANE(contours[0]), &clipc) == 1)   {
	  draw_contour (clipc, 0, 0, 1, RGB_BLUE, RGB_YELLOW);
	  discard_contour (&clipc);
	}
#endif

	// graph the solid angle of contour 0 as seen from nc-1
	// first we must subsample 1 somehow... in its orthobasis?
	graph_solidangle (contours[0], &contours[1], ncontours-2, contours[ncontours-1], 
					  r_density, dirty_data, RGB_GREEN);
	dirty_data = FALSE;
  }
  else if (ncontours > 0) {
	/* assemble contours into list */
	clist = create_list (sizeof (CONTOUR *));

	for (k = 0; k < ncontours; k++)
	  insert_list (clist, (void *)contours[k]);

	/* experiment: draw exploded contours */

	if (dirty_data) {
	  if (events != NULL)
		apen_discard_event_list (&events);
	  get_exploded_contour_events (clist, &events);
	  dirty_data = FALSE;
	}

	draw_exploded_contours (clist);
	draw_antipenumbral_events (clist, events, event_types, NULL, control);
  }

  /* incident lines */
  if (ncontours == 0 && nsegments == 4)   {
	if (draw_cylinders)
	  push_lighting(1);

	/* disable highlighting */
	/* pref_seg = -1; */

	nsolns = draw_incident_lines ();
	draw_segments (FALSE, 0, 3, RGB_GREEN, RGB_RED);

	if (draw_cylinders)
	  pop_lighting();
  }

  /* incident lines */
  if (ncontours == 0 && nsegments == 3)   {
	if (draw_cylinders)
	  push_lighting(1);

	if (show_implicit_quadric && nsegments == 3)
	  draw_implicit_regulus ((POINT *)&segments[0][0], (POINT *)&segments[0][4],
							 (POINT *)&segments[1][0], (POINT *)&segments[1][4],
							 (POINT *)&segments[2][0], (POINT *)&segments[2][4]);

	if (draw_cylinders)
	  pop_lighting();
  }

  /* plucker dot product */
  if (ncontours == 0 && nsegments == 2)   {
	if (draw_cylinders)
	  push_lighting(1);

	pref_seg = -1;

	draw_segments (FALSE, 0, 0, RGB_RED, RGB_CYAN);
	draw_segments (FALSE, 1, 1, RGB_GREEN,   RGB_CYAN);

	draw_twoline_distance (RGB_GREEN, RGB_YELLOW, RGB_RED);

	if (draw_cylinders)
	  pop_lighting();
  }

  /* general regulus */
  if (ncontours == 0 && regulus && nsegments == 3)	{
	push_lighting(1);
	MPACK (RGB_CYAN);

	if (r_density > 0)
	  draw_generic_regulus ((POINT *)&segments[0][0], (POINT *)&segments[0][4],
							(POINT *)&segments[1][0], (POINT *)&segments[1][4],
							(POINT *)&segments[2][0], (POINT *)&segments[2][4],
							TRUE,
							r_density,
							0.8,
							shaded,
							RGB_CYAN,
							FALSE);

	MPACK (RGB_RED);
	if (animating)
	  draw_parametrized_regulus ((POINT *)&segments[0][0], (POINT *)&segments[0][4],
								 (POINT *)&segments[1][0], (POINT *)&segments[1][4],
								 (POINT *)&segments[2][0], (POINT *)&segments[2][4],
								 time_val, time_val,
								 TRUE,
								 r_density,
								 1.2,
								 shaded,
								 RGB_RED,
								 FALSE);

	draw_segments (FALSE, 0, nsegments-1, RGB_GREEN, RGB_GREEN);
	pop_lighting();
  }

  if (nsegments == 3 && show_containment) {
	push_lighting (1);
	draw_containment_lines ();
	pop_lighting();
  }

report:

#ifdef XXX_UNKNOWN_BUG
  if (tape_getmode() != TAPE_PLAYBACK && !regulus && !printscreen)   {
	set_orthoviewing();

	CPACK (RGB_ABYSS);
	sprintf (str2, "%d array bytes extant\n", array_bytes_extant); 
	cmov2i (2, 2);
	charstr (str2);

	if (nsolns == -1)
	  CPACK (RGB_ORANGE);
	else 
	  CPACK (nsolns == 0 ? RGB_RED : RGB_GREEN);

	if (experimental)   {
	  CPACK (RGB_GREEN);
	  sprintf (str, "   PVE %2d, IVE %2d, UVE %2d, PEEE %2d, IEEE %2d EEEE %2d\n",
			   event_types & PVE_EVENT_TYPE,
			   event_types & IVE_EVENT_TYPE,
			   event_types & UVE_EVENT_TYPE,
			   event_types & PEEE_EVENT_TYPE,
			   event_types & IEEE_EVENT_TYPE,
			   event_types & EEEE_EVENT_TYPE);
	  cmov2i (2, 14);
	  charstr (str);

	  CPACK (RGB_ABYSS);
	  sprintf (str, "[%4d events] %4d PVE, %4d IVE, %4d UVE, %4d PEEE, %4d IEEE, %4d EEEE, %4d unknown\n",
			   LNULL_OR_LENGTH(events),
			   count_events (events, PVE_EVENT_TYPE),
			   count_events (events, IVE_EVENT_TYPE),
			   count_events (events, UVE_EVENT_TYPE),
			   count_events (events, PEEE_EVENT_TYPE),
			   count_events (events, IEEE_EVENT_TYPE),
			   count_events (events, EEEE_EVENT_TYPE),
			   count_events (events, UNKNOWN_EVENT_TYPE));
	  cmov2i (2, 26);
	  charstr (str);

	  CPACK (RGB_GREEN);
	  sprintf (str, "n = %4d: %4d vertices, %4d facets, %4d ridges, %4d cliffs\n", ps, nv, nf, nr, nc);
	  cmov2i (2, 38);
	  charstr (str);
	}
	else if (do_blockers)	{
	  CPACK (RGB_YELLOW);
	  sprintf (str, "   vistype is %s", vistypename[vis]);
	  cmov2i (2, 18);
	  charstr (str);
    
	  if (do_blocker_type)    {
		readsource (SRC_BACK);
		rectzoom (4.0, 4.0);
		rectcopy (10, 14, 250, 30, 10, 50);
		rectzoom (1.0, 1.0);
	  }
	}
	else if (do_ff)	{
	  CPACK (RGB_YELLOW);
	  sprintf (str, "   form factor is %10.6lf", form_factor);
	  cmov2i (2, 18);
	  charstr (str);
	}
  }

  if (!experimental)
	dirty_data = FALSE;

  if (spinning)
	dirty_data = TRUE;

  {
    extern char *script_playback_file;
	
    if (script_playback_file)
	  dirty_data = TRUE;
  }
#endif

}

int
stoprend()
{
  if (tape_getmode() == TAPE_PLAYBACK)
	return (0);
  return (getvaluator (LEFTMOUSE) || getvaluator (MIDDLEMOUSE) || getvaluator (RIGHTMOUSE));
}

/* master draw routine */
void
REFRESH( void )
{
  if (!accumulating 
	  || (!script_playback_file && (leftdown || middledown || rightdown))
	  || (script_playback_file && (frame_num < first_frame || frame_num > last_frame)))	{
	REFRESH_ns();
	my_swapbuffers();
  }
  else    {
	doaccbuf (REFRESH_ns, stoprend, my_swapbuffers);
  }

  if (animating)  {
	if ((time_val += time_step) > 1.0)  {
	  time_val = 1.0 - (time_val - 1.0);
	  time_step = -time_step;
	}
	else if (time_val < 0.0)    {
	  time_val = -time_val;
	  time_step = -time_step;
	}
	dirty_data = TRUE;
  }
}
